# Copyright (C) 2011, 2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""bzr-upload command implementations."""

from ... import commands, config, errors, lazy_import, option, osutils

lazy_import.lazy_import(globals(), """
import stat

from breezy import (
    controldir,
    globbing,
    ignores,
    revision,
    transport,
    urlutils,
    )
""")


auto_option = config.Option(
    'upload_auto', default=False, from_unicode=config.bool_from_store,
    help="""\
Whether upload should occur when the tip of the branch changes.
""")
auto_quiet_option = config.Option(
    'upload_auto_quiet', default=False, from_unicode=config.bool_from_store,
    help="""\
Whether upload should occur quietly.
""")
location_option = config.Option(
    'upload_location', default=None,
    help="""\
The url to upload the working tree to.
""")
revid_location_option = config.Option(
    'upload_revid_location', default='.bzr-upload.revid',
    help="""\
The relative path to be used to store the uploaded revid.

The only bzr-related info uploaded with the working tree is the corresponding
revision id. The uploaded working tree is not linked to any other bzr data.

If the layout of your remote server is such that you can't write in the
root directory but only in the directories inside that root, you will need
to use the 'upload_revid_location' configuration variable to specify the
relative path to be used. That configuration variable can be specified in
locations.conf or branch.conf.

For example, given the following layout:

  Project/
    private/
    public/

you may have write access in 'private' and 'public' but in 'Project'
itself. In that case, you can add the following in your locations.conf or
branch.conf file:

  upload_revid_location = private/.bzr-upload.revid
""")


# FIXME: Add more tests around invalid paths or relative paths that doesn't
# exist on remote (if only to get proper error messages) for
# 'upload_revid_location'


class BzrUploader:

    def __init__(self, branch, to_transport, outf, tree, rev_id,
                 quiet=False):
        self.branch = branch
        self.to_transport = to_transport
        self.outf = outf
        self.tree = tree
        self.rev_id = rev_id
        self.quiet = quiet
        self._pending_deletions = []
        self._pending_renames = []
        self._uploaded_revid = None
        self._ignored = None

    def _up_stat(self, relpath):
        return self.to_transport.stat(urlutils.escape(relpath))

    def _up_rename(self, old_path, new_path):
        return self.to_transport.rename(urlutils.escape(old_path),
                                        urlutils.escape(new_path))

    def _up_delete(self, relpath):
        return self.to_transport.delete(urlutils.escape(relpath))

    def _up_delete_tree(self, relpath):
        return self.to_transport.delete_tree(urlutils.escape(relpath))

    def _up_mkdir(self, relpath, mode):
        return self.to_transport.mkdir(urlutils.escape(relpath), mode)

    def _up_rmdir(self, relpath):
        return self.to_transport.rmdir(urlutils.escape(relpath))

    def _up_put_bytes(self, relpath, bytes, mode):
        self.to_transport.put_bytes(urlutils.escape(relpath), bytes, mode)

    def _up_get_bytes(self, relpath):
        return self.to_transport.get_bytes(urlutils.escape(relpath))

    def set_uploaded_revid(self, rev_id):
        # XXX: Add tests for concurrent updates, etc.
        revid_path = self.branch.get_config_stack().get('upload_revid_location')
        self.to_transport.put_bytes(urlutils.escape(revid_path), rev_id)
        self._uploaded_revid = rev_id

    def get_uploaded_revid(self):
        if self._uploaded_revid is None:
            revid_path = self.branch.get_config_stack(
                ).get('upload_revid_location')
            try:
                self._uploaded_revid = self._up_get_bytes(revid_path)
            except transport.NoSuchFile:
                # We have not uploaded to here.
                self._uploaded_revid = revision.NULL_REVISION
        return self._uploaded_revid

    def _get_ignored(self):
        if self._ignored is None:
            try:
                ignore_file_path = '.bzrignore-upload'
                ignore_file = self.tree.get_file(ignore_file_path)
            except transport.NoSuchFile:
                ignored_patterns = []
            else:
                ignored_patterns = ignores.parse_ignore_file(ignore_file)
            self._ignored = globbing.Globster(ignored_patterns)
        return self._ignored

    def is_ignored(self, relpath):
        glob = self._get_ignored()
        ignored = glob.match(relpath)
        import os
        if not ignored:
            # We still need to check that all parents are not ignored
            dir = os.path.dirname(relpath)
            while dir and not ignored:
                ignored = glob.match(dir)
                if not ignored:
                    dir = os.path.dirname(dir)
        return ignored

    def upload_file(self, old_relpath, new_relpath, mode=None):
        if mode is None:
            if self.tree.is_executable(new_relpath):
                mode = 0o775
            else:
                mode = 0o664
        if not self.quiet:
            self.outf.write('Uploading %s\n' % old_relpath)
        self._up_put_bytes(
            old_relpath, self.tree.get_file_text(new_relpath), mode)

    def _force_clear(self, relpath):
        try:
            st = self._up_stat(relpath)
            if stat.S_ISDIR(st.st_mode):
                # A simple rmdir may not be enough
                if not self.quiet:
                    self.outf.write('Clearing {}/{}\n'.format(
                        self.to_transport.external_url(), relpath))
                self._up_delete_tree(relpath)
            elif stat.S_ISLNK(st.st_mode):
                if not self.quiet:
                    self.outf.write('Clearing {}/{}\n'.format(
                        self.to_transport.external_url(), relpath))
                self._up_delete(relpath)
        except errors.PathError:
            pass

    def upload_file_robustly(self, relpath, mode=None):
        """Upload a file, clearing the way on the remote side.

        When doing a full upload, it may happen that a directory exists where
        we want to put our file.
        """
        self._force_clear(relpath)
        self.upload_file(relpath, relpath, mode)

    def upload_symlink(self, relpath, target):
        self.to_transport.symlink(target, relpath)

    def upload_symlink_robustly(self, relpath, target):
        """Handle uploading symlinks.
        """
        self._force_clear(relpath)
        # Target might not be there at this time; dummy file should be
        # overwritten at some point, possibly by another upload.
        target = osutils.normpath(osutils.pathjoin(
            osutils.dirname(relpath),
            target)
        )
        self.upload_symlink(relpath, target)

    def make_remote_dir(self, relpath, mode=None):
        if mode is None:
            mode = 0o775
        self._up_mkdir(relpath, mode)

    def make_remote_dir_robustly(self, relpath, mode=None):
        """Create a remote directory, clearing the way on the remote side.

        When doing a full upload, it may happen that a file exists where we
        want to create our directory.
        """
        try:
            st = self._up_stat(relpath)
            if not stat.S_ISDIR(st.st_mode):
                if not self.quiet:
                    self.outf.write('Deleting {}/{}\n'.format(
                        self.to_transport.external_url(), relpath))
                self._up_delete(relpath)
            else:
                # Ok the remote dir already exists, nothing to do
                return
        except errors.PathError:
            pass
        self.make_remote_dir(relpath, mode)

    def delete_remote_file(self, relpath):
        if not self.quiet:
            self.outf.write('Deleting %s\n' % relpath)
        self._up_delete(relpath)

    def delete_remote_dir(self, relpath):
        if not self.quiet:
            self.outf.write('Deleting %s\n' % relpath)
        self._up_rmdir(relpath)
        # XXX: Add a test where a subdir is ignored but we still want to
        # delete the dir -- vila 100106

    def delete_remote_dir_maybe(self, relpath):
        """Try to delete relpath, keeping failures to retry later."""
        try:
            self._up_rmdir(relpath)
        # any kind of PathError would be OK, though we normally expect
        # DirectoryNotEmpty
        except errors.PathError:
            self._pending_deletions.append(relpath)

    def finish_deletions(self):
        if self._pending_deletions:
            # Process the previously failed deletions in reverse order to
            # delete children before parents
            for relpath in reversed(self._pending_deletions):
                self._up_rmdir(relpath)
            # The following shouldn't be needed since we use it once per
            # upload, but better safe than sorry ;-)
            self._pending_deletions = []

    def rename_remote(self, old_relpath, new_relpath):
        """Rename a remote file or directory taking care of collisions.

        To avoid collisions during bulk renames, each renamed target is
        temporarily assigned a unique name. When all renames have been done,
        each target get its proper name.
        """
        # We generate a sufficiently random name to *assume* that
        # no collisions will occur and don't worry about it (nor
        # handle it).
        import os
        import random
        import time

        stamp = '.tmp.%.9f.%d.%d' % (time.time(),
                                     os.getpid(),
                                     random.randint(0, 0x7FFFFFFF))
        if not self.quiet:
            self.outf.write('Renaming {} to {}\n'.format(old_relpath, new_relpath))
        self._up_rename(old_relpath, stamp)
        self._pending_renames.append((stamp, new_relpath))

    def finish_renames(self):
        for (stamp, new_path) in self._pending_renames:
            self._up_rename(stamp, new_path)
        # The following shouldn't be needed since we use it once per upload,
        # but better safe than sorry ;-)
        self._pending_renames = []

    def upload_full_tree(self):
        self.to_transport.ensure_base()  # XXX: Handle errors (add
        # --create-prefix option ?)
        with self.tree.lock_read():
            for relpath, ie in self.tree.iter_entries_by_dir():
                if relpath in ('', '.bzrignore', '.bzrignore-upload'):
                    # skip root ('')
                    # .bzrignore and .bzrignore-upload have no meaning outside
                    # a working tree so do not upload them
                    continue
                if self.is_ignored(relpath):
                    if not self.quiet:
                        self.outf.write('Ignoring %s\n' % relpath)
                    continue
                if ie.kind == 'file':
                    self.upload_file_robustly(relpath)
                elif ie.kind == 'symlink':
                    try:
                        self.upload_symlink_robustly(
                            relpath, ie.symlink_target)
                    except errors.TransportNotPossible:
                        if not self.quiet:
                            target = self.tree.path_content_summary(relpath)[3]
                            self.outf.write('Not uploading symlink %s -> %s\n'
                                            % (relpath, target))
                elif ie.kind == 'directory':
                    self.make_remote_dir_robustly(relpath)
                else:
                    raise NotImplementedError
            self.set_uploaded_revid(self.rev_id)

    def upload_tree(self):
        # If we can't find the revid file on the remote location, upload the
        # full tree instead
        rev_id = self.get_uploaded_revid()

        if rev_id == revision.NULL_REVISION:
            if not self.quiet:
                self.outf.write('No uploaded revision id found,'
                                ' switching to full upload\n')
            self.upload_full_tree()
            # We're done
            return

        # Check if the revision hasn't already been uploaded
        if rev_id == self.rev_id:
            if not self.quiet:
                self.outf.write('Remote location already up to date\n')

        from_tree = self.branch.repository.revision_tree(rev_id)
        self.to_transport.ensure_base()  # XXX: Handle errors (add
        # --create-prefix option ?)
        changes = self.tree.changes_from(from_tree)
        with self.tree.lock_read():
            for change in changes.removed:
                if self.is_ignored(change.path[0]):
                    if not self.quiet:
                        self.outf.write('Ignoring %s\n' % change.path[0])
                    continue
                if change.kind[0] == 'file':
                    self.delete_remote_file(change.path[0])
                elif change.kind[0] == 'directory':
                    self.delete_remote_dir_maybe(change.path[0])
                elif change.kind[0] == 'symlink':
                    self.delete_remote_file(change.path[0])
                else:
                    raise NotImplementedError

            for change in changes.renamed:
                if self.is_ignored(change.path[0]) and self.is_ignored(change.path[1]):
                    if not self.quiet:
                        self.outf.write('Ignoring %s\n' % change.path[0])
                        self.outf.write('Ignoring %s\n' % change.path[1])
                    continue
                if change.changed_content:
                    # We update the change.path[0] content because renames and
                    # deletions are differed.
                    self.upload_file(change.path[0], change.path[1])
                self.rename_remote(change.path[0], change.path[1])
            self.finish_renames()
            self.finish_deletions()

            for change in changes.kind_changed:
                if self.is_ignored(change.path[1]):
                    if not self.quiet:
                        self.outf.write('Ignoring %s\n' % change.path[1])
                    continue
                if change.kind[0] in ('file', 'symlink'):
                    self.delete_remote_file(change.path[0])
                elif change.kind[0] == 'directory':
                    self.delete_remote_dir(change.path[0])
                else:
                    raise NotImplementedError

                if change.kind[1] == 'file':
                    self.upload_file(change.path[1], change.path[1])
                elif change.kind[1] == 'symlink':
                    target = self.tree.get_symlink_target(change.path[1])
                    self.upload_symlink(change.path[1], target)
                elif change.kind[1] == 'directory':
                    self.make_remote_dir(change.path[1])
                else:
                    raise NotImplementedError

            for change in changes.added + changes.copied:
                if self.is_ignored(change.path[1]):
                    if not self.quiet:
                        self.outf.write('Ignoring %s\n' % change.path[1])
                    continue
                if change.kind[1] == 'file':
                    self.upload_file(change.path[1], change.path[1])
                elif change.kind[1] == 'directory':
                    self.make_remote_dir(change.path[1])
                elif change.kind[1] == 'symlink':
                    target = self.tree.get_symlink_target(change.path[1])
                    try:
                        self.upload_symlink(change.path[1], target)
                    except errors.TransportNotPossible:
                        if not self.quiet:
                            self.outf.write('Not uploading symlink %s -> %s\n'
                                            % (change.path[1], target))
                else:
                    raise NotImplementedError

            # XXX: Add a test for exec_change
            for change in changes.modified:
                if self.is_ignored(change.path[1]):
                    if not self.quiet:
                        self.outf.write('Ignoring %s\n' % change.path[1])
                    continue
                if change.kind[1] == 'file':
                    self.upload_file(change.path[1], change.path[1])
                elif change.kind[1] == 'symlink':
                    target = self.tree.get_symlink_target(change.path[1])
                    self.upload_symlink(change.path[1], target)
                else:
                    raise NotImplementedError

            self.set_uploaded_revid(self.rev_id)


class CannotUploadToWorkingTree(errors.CommandError):

    _fmt = 'Cannot upload to a bzr managed working tree: %(url)s".'


class DivergedUploadedTree(errors.CommandError):

    _fmt = ("Your branch (%(revid)s)"
            " and the uploaded tree (%(uploaded_revid)s) have diverged: ")


class cmd_upload(commands.Command):
    """Upload a working tree, as a whole or incrementally.

    If no destination is specified use the last one used.
    If no revision is specified upload the changes since the last upload.

    Changes include files added, renamed, modified or removed.
    """
    _see_also = ['plugins/upload']
    takes_args = ['location?']
    takes_options = [
        'revision',
        'remember',
        'overwrite',
        option.Option('full', 'Upload the full working tree.'),
        option.Option('quiet', 'Do not output what is being done.',
                      short_name='q'),
        option.Option('directory',
                      help='Branch to upload from, '
                      'rather than the one containing the working directory.',
                      short_name='d',
                      type=str,
                      ),
        option.Option('auto',
                      'Trigger an upload from this branch whenever the tip '
                      'revision changes.')
        ]

    def run(self, location=None, full=False, revision=None, remember=None,
            directory=None, quiet=False, auto=None, overwrite=False
            ):
        if directory is None:
            directory = '.'

        (wt, branch,
         relpath) = controldir.ControlDir.open_containing_tree_or_branch(
             directory)

        if wt:
            locked = wt
        else:
            locked = branch
        with locked.lock_read():
            if wt:
                changes = wt.changes_from(wt.basis_tree())

                if revision is None and changes.has_changed():
                    raise errors.UncommittedChanges(wt)

            conf = branch.get_config_stack()
            if location is None:
                stored_loc = conf.get('upload_location')
                if stored_loc is None:
                    raise errors.CommandError(
                        'No upload location known or specified.')
                else:
                    # FIXME: Not currently tested
                    display_url = urlutils.unescape_for_display(stored_loc,
                                                                self.outf.encoding)
                    self.outf.write("Using saved location: %s\n" % display_url)
                    location = stored_loc

            to_transport = transport.get_transport(location)

            # Check that we are not uploading to a existing working tree.
            try:
                to_bzr_dir = controldir.ControlDir.open_from_transport(
                    to_transport)
                has_wt = to_bzr_dir.has_workingtree()
            except errors.NotBranchError:
                has_wt = False
            except errors.NotLocalUrl:
                # The exception raised is a bit weird... but that's life.
                has_wt = True

            if has_wt:
                raise CannotUploadToWorkingTree(url=location)
            if revision is None:
                rev_id = branch.last_revision()
            else:
                if len(revision) != 1:
                    raise errors.CommandError(
                        'bzr upload --revision takes exactly 1 argument')
                rev_id = revision[0].in_history(branch).rev_id

            tree = branch.repository.revision_tree(rev_id)

            uploader = BzrUploader(branch, to_transport, self.outf, tree,
                                   rev_id, quiet=quiet)

            if not overwrite:
                prev_uploaded_rev_id = uploader.get_uploaded_revid()
                graph = branch.repository.get_graph()
                if not graph.is_ancestor(prev_uploaded_rev_id, rev_id):
                    raise DivergedUploadedTree(
                        revid=rev_id, uploaded_revid=prev_uploaded_rev_id)
            if full:
                uploader.upload_full_tree()
            else:
                uploader.upload_tree()

        # We uploaded successfully, remember it
        with branch.lock_write():
            upload_location = conf.get('upload_location')
            if upload_location is None or remember:
                conf.set('upload_location',
                         urlutils.unescape(to_transport.base))
            if auto is not None:
                conf.set('upload_auto', auto)



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ... import commands, controldir, errors


class cmd_file_refs(commands.Command):
    __doc__ = \
        """Find the inventories that reference a particular version of a text."""

    hidden = True
    takes_args = ['file_id', 'rev_id']
    takes_options = ['directory']

    def run(self, file_id, rev_id, directory='.'):
        file_id = file_id.encode()
        rev_id = rev_id.encode()
        bd, relpath = controldir.ControlDir.open_containing(directory)
        repo = bd.find_repository()
        self.add_cleanup(repo.lock_read().unlock)
        inv_vf = repo.inventories
        all_invs = [k[-1] for k in inv_vf.keys()]
        # print len(all_invs)
        for inv in repo.iter_inventories(all_invs, 'unordered'):
            try:
                entry = inv.get_entry(file_id)
            except errors.NoSuchId:
                # This file doesn't even appear in this inv.
                continue
            if entry.revision == rev_id:
                self.outf.write(inv.revision_id + b'\n')



# Copyright (C) 2006-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy import errors, tests, workingtree
from breezy.mutabletree import BadReferenceTarget
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestBasisInventory(TestCaseWithWorkingTree):

    def make_trees(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/file1'])
        tree.add('file1')
        sub_tree = self.make_branch_and_tree('tree/sub-tree')
        sub_tree.commit('commit')
        return tree, sub_tree

    def _references_unsupported(self, tree):
        if not tree.supports_tree_reference():
            raise tests.TestNotApplicable(
                'Tree format does not support references')
        else:
            self.fail('%r does not support references but should'
                      % (tree, ))

    def make_nested_trees(self):
        tree, sub_tree = self.make_trees()
        try:
            tree.add_reference(sub_tree)
        except errors.UnsupportedOperation:
            self._references_unsupported(tree)
        return tree, sub_tree

    def test_add_reference(self):
        tree, sub_tree = self.make_nested_trees()
        with tree.lock_write():
            if tree.supports_setting_file_ids():
                sub_tree_root_id = sub_tree.path2id('')
                self.assertEqual(tree.path2id('sub-tree'), sub_tree_root_id)
            self.assertEqual(tree.kind('sub-tree'), 'tree-reference')
            tree.commit('commit reference')
            basis = tree.basis_tree()
            with basis.lock_read():
                sub_tree = tree.get_nested_tree('sub-tree')
                self.assertEqual(
                    sub_tree.last_revision(),
                    tree.get_reference_revision('sub-tree'))
        self.assertEqual(['sub-tree'], list(tree.iter_references()))

    def test_add_reference_same_root(self):
        tree = self.make_branch_and_tree('tree')
        if not tree.supports_setting_file_ids():
            self.skipTest('format does not support setting file ids')
        self.build_tree(['tree/file1'])
        tree.add('file1')
        tree.set_root_id(b'root-id')
        sub_tree = self.make_branch_and_tree('tree/sub-tree')
        sub_tree.set_root_id(b'root-id')
        try:
            self.assertRaises(BadReferenceTarget, tree.add_reference,
                              sub_tree)
        except errors.UnsupportedOperation:
            self._references_unsupported(tree)

    def test_root_present(self):
        """Subtree root is present, though not the working tree root"""
        tree, sub_tree = self.make_trees()
        if not tree.supports_setting_file_ids():
            self.skipTest('format does not support setting file ids')
        sub_tree.set_root_id(tree.path2id('file1'))
        try:
            self.assertRaises(BadReferenceTarget, tree.add_reference,
                              sub_tree)
        except errors.UnsupportedOperation:
            self._references_unsupported(tree)

    def test_add_non_subtree(self):
        tree, sub_tree = self.make_trees()
        os.rename('tree/sub-tree', 'sibling')
        sibling = workingtree.WorkingTree.open('sibling')
        try:
            self.assertRaises(BadReferenceTarget, tree.add_reference,
                              sibling)
        except errors.UnsupportedOperation:
            self._references_unsupported(tree)

    def test_get_nested_tree(self):
        tree, sub_tree = self.make_nested_trees()
        with tree.lock_read():
            sub_tree2 = tree.get_nested_tree('sub-tree')
            self.assertEqual(sub_tree.basedir, sub_tree2.basedir)

    def test_get_containing_nested_tree(self):
        tree, sub_tree = self.make_nested_trees()
        self.build_tree_contents([('tree/sub-tree/foo', 'bar')])
        sub_tree.add('foo')
        sub_tree.commit('rev1')
        with tree.lock_read():
            sub_tree2, subpath = tree.get_containing_nested_tree(
                'sub-tree/foo')
            self.assertEqual(sub_tree.basedir, sub_tree2.basedir)
            self.assertEqual(subpath, 'foo')
            self.assertEqual((None, None), tree.get_containing_nested_tree(
                'not-subtree/bar'))



# Copyright (C) 2006-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Whitebox tests for annotate functionality."""

import codecs
from io import BytesIO, StringIO

from .. import annotate, tests
from .ui_testing import StringIOWithEncoding


def annotation(text):
    return [tuple(l.split(b' ', 1)) for l in text.splitlines(True)]


parent_1 = annotation(b"""\
rev1 a
rev2 b
rev3 c
rev4 d
rev5 e
""")


parent_2 = annotation(b"""\
rev1 a
rev3 c
rev4 d
rev6 f
rev7 e
rev8 h
""")


expected_2_1 = annotation(b"""\
rev1 a
blahblah b
rev3 c
rev4 d
rev7 e
""")


# a: in both, same value, kept
# b: in 1, kept
# c: in both, same value, kept
# d: in both, same value, kept
# e: 1 and 2 disagree, so it goes to blahblah
# f: in 2, but not in new, so ignored
# g: not in 1 or 2, so it goes to blahblah
# h: only in parent 2, so 2 gets it
expected_1_2_2 = annotation(b"""\
rev1 a
rev2 b
rev3 c
rev4 d
blahblah e
blahblah g
rev8 h
""")


new_1 = b"""\
a
b
c
d
e
""".splitlines(True)

expected_1 = annotation(b"""\
blahblah a
blahblah b
blahblah c
blahblah d
blahblah e
""")


new_2 = b"""\
a
b
c
d
e
g
h
""".splitlines(True)


# For the 'duplicate' series, both sides introduce the same change, which then
# gets merged around. The last-modified should properly reflect this.
# We always change the fourth line so that the file is properly tracked as
# being modified in each revision. In reality, this probably would happen over
# many revisions, and it would be a different line that changes.
# BASE
#  |\
#  A B  # line should be annotated as new for A and B
#  |\|
#  C D  # line should 'converge' and say A
#  |/
#  E    # D should supersede A and stay as D (not become E because C references
#         A)
duplicate_base = annotation(b"""\
rev-base first
rev-base second
rev-base third
rev-base fourth-base
""")

duplicate_A = annotation(b"""\
rev-base first
rev-A alt-second
rev-base third
rev-A fourth-A
""")

duplicate_B = annotation(b"""\
rev-base first
rev-B alt-second
rev-base third
rev-B fourth-B
""")

duplicate_C = annotation(b"""\
rev-base first
rev-A alt-second
rev-base third
rev-C fourth-C
""")

duplicate_D = annotation(b"""\
rev-base first
rev-A alt-second
rev-base third
rev-D fourth-D
""")

duplicate_E = annotation(b"""\
rev-base first
rev-A alt-second
rev-base third
rev-E fourth-E
""")


class TestAnnotate(tests.TestCaseWithTransport):

    def create_merged_trees(self):
        """create 2 trees with merges between them.

        rev-1 --+
         |      |
        rev-2  rev-1_1_1
         |      |
         +------+
         |
        rev-3
        """
        builder = self.make_branch_builder('branch')
        builder.start_series()
        self.addCleanup(builder.finish_series)
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('a', b'a-id', 'file', b'first\n')),
            ], timestamp=1166046000.00, timezone=0, committer="joe@foo.com",
            revision_id=b'rev-1')
        builder.build_snapshot([b'rev-1'], [
            ('modify', ('a', b'first\nsecond\n')),
            ], timestamp=1166046001.00, timezone=0, committer="joe@foo.com",
            revision_id=b'rev-2')
        builder.build_snapshot([b'rev-1'], [
            ('modify', ('a', b'first\nthird\n')),
            ], timestamp=1166046002.00, timezone=0, committer="barry@foo.com",
            revision_id=b'rev-1_1_1')
        builder.build_snapshot([b'rev-2', b'rev-1_1_1'], [
            ('modify', ('a', b'first\nsecond\nthird\n')),
            ], timestamp=1166046003.00, timezone=0, committer="sal@foo.com",
            revision_id=b'rev-3')
        return builder

    def create_deeply_merged_trees(self):
        """Create some trees with a more complex merge history.

        rev-1 --+
         |      |
        rev-2  rev-1_1_1 --+
         |      |          |
         +------+          |
         |      |          |
        rev-3  rev-1_1_2  rev-1_2_1 ------+
         |      |          |              |
         +------+          |              |
         |                 |              |
        rev-4             rev-1_2_2  rev-1_3_1
         |                 |              |
         +-----------------+              |
         |                                |
        rev-5                             |
         |                                |
         +--------------------------------+
         |
        rev-6
        """
        builder = self.create_merged_trees()
        builder.build_snapshot([b'rev-1_1_1'], [], revision_id=b'rev-1_1_2')
        builder.build_snapshot([b'rev-3', b'rev-1_1_2'],
                               [], revision_id=b'rev-4')
        builder.build_snapshot([b'rev-1_1_1'], [
            ('modify', ('a', b'first\nthird\nfourth\n')),
            ], timestamp=1166046003.00, timezone=0, committer="jerry@foo.com",
            revision_id=b'rev-1_2_1')
        builder.build_snapshot([b'rev-1_2_1'], [],
                               timestamp=1166046004.00, timezone=0, committer="jerry@foo.com",
                               revision_id=b'rev-1_2_2')
        builder.build_snapshot([b'rev-4', b'rev-1_2_2'], [
            ('modify', ('a', b'first\nsecond\nthird\nfourth\n')),
            ], timestamp=1166046004.00, timezone=0, committer="jerry@foo.com",
            revision_id=b'rev-5')
        builder.build_snapshot([b'rev-1_2_1'], [
            ('modify', ('a', b'first\nthird\nfourth\nfifth\nsixth\n')),
            ], timestamp=1166046005.00, timezone=0, committer="george@foo.com",
            revision_id=b'rev-1_3_1')
        builder.build_snapshot([b'rev-5', b'rev-1_3_1'], [
            ('modify', ('a',
                        b'first\nsecond\nthird\nfourth\nfifth\nsixth\n')),
            ], revision_id=b'rev-6')
        return builder

    def create_duplicate_lines_tree(self):
        builder = self.make_branch_builder('branch')
        builder.start_series()
        self.addCleanup(builder.finish_series)
        base_text = b''.join(l for r, l in duplicate_base)
        a_text = b''.join(l for r, l in duplicate_A)
        b_text = b''.join(l for r, l in duplicate_B)
        c_text = b''.join(l for r, l in duplicate_C)
        d_text = b''.join(l for r, l in duplicate_D)
        e_text = b''.join(l for r, l in duplicate_E)
        builder.build_snapshot(None, [
            ('add', ('', b'root-id', 'directory', None)),
            ('add', ('file', b'file-id', 'file', base_text)),
            ], revision_id=b'rev-base')
        builder.build_snapshot([b'rev-base'], [
            ('modify', ('file', a_text))],
            revision_id=b'rev-A')
        builder.build_snapshot([b'rev-base'], [
            ('modify', ('file', b_text))],
            revision_id=b'rev-B')
        builder.build_snapshot([b'rev-A'], [
            ('modify', ('file', c_text))],
            revision_id=b'rev-C')
        builder.build_snapshot([b'rev-B', b'rev-A'], [
            ('modify', ('file', d_text))],
            revision_id=b'rev-D')
        builder.build_snapshot([b'rev-C', b'rev-D'], [
            ('modify', ('file', e_text))],
            revision_id=b'rev-E')
        return builder

    def assertAnnotateEqualDiff(self, actual, expected):
        if actual != expected:
            # Create an easier to understand diff when the lines don't actually
            # match
            self.assertEqualDiff(''.join('\t'.join(l) for l in expected),
                                 ''.join('\t'.join(l) for l in actual))

    def assertBranchAnnotate(self, expected, branch, path, revision_id,
                             verbose=False, full=False, show_ids=False):
        tree = branch.repository.revision_tree(revision_id)
        to_file = StringIO()
        annotate.annotate_file_tree(tree, path, to_file,
                                    verbose=verbose, full=full, show_ids=show_ids, branch=branch)
        self.assertAnnotateEqualDiff(to_file.getvalue(), expected)

    def assertRepoAnnotate(self, expected, repo, path, revision_id):
        """Assert that the revision is properly annotated."""
        actual = list(repo.revision_tree(revision_id).annotate_iter(path))
        self.assertAnnotateEqualDiff(actual, expected)

    def test_annotate_duplicate_lines(self):
        # XXX: Should this be a per_repository test?
        builder = self.create_duplicate_lines_tree()
        repo = builder.get_branch().repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertRepoAnnotate(duplicate_base, repo, 'file', b'rev-base')
        self.assertRepoAnnotate(duplicate_A, repo, 'file', b'rev-A')
        self.assertRepoAnnotate(duplicate_B, repo, 'file', b'rev-B')
        self.assertRepoAnnotate(duplicate_C, repo, 'file', b'rev-C')
        self.assertRepoAnnotate(duplicate_D, repo, 'file', b'rev-D')
        self.assertRepoAnnotate(duplicate_E, repo, 'file', b'rev-E')

    def test_annotate_shows_dotted_revnos(self):
        builder = self.create_merged_trees()

        self.assertBranchAnnotate('1     joe@foo | first\n'
                                  '2     joe@foo | second\n'
                                  '1.1.1 barry@f | third\n',
                                  builder.get_branch(), 'a', b'rev-3')

    def test_annotate_limits_dotted_revnos(self):
        """Annotate should limit dotted revnos to a depth of 12"""
        builder = self.create_deeply_merged_trees()

        self.assertBranchAnnotate('1     joe@foo | first\n'
                                  '2     joe@foo | second\n'
                                  '1.1.1 barry@f | third\n'
                                  '1.2.1 jerry@f | fourth\n'
                                  '1.3.1 george@ | fifth\n'
                                  '              | sixth\n',
                                  builder.get_branch(), 'a', b'rev-6',
                                  verbose=False, full=False)

        self.assertBranchAnnotate('1     joe@foo | first\n'
                                  '2     joe@foo | second\n'
                                  '1.1.1 barry@f | third\n'
                                  '1.2.1 jerry@f | fourth\n'
                                  '1.3.1 george@ | fifth\n'
                                  '1.3.1 george@ | sixth\n',
                                  builder.get_branch(), 'a', b'rev-6',
                                  verbose=False, full=True)

        # verbose=True shows everything, the full revno, user id, and date
        self.assertBranchAnnotate('1     joe@foo.com    20061213 | first\n'
                                  '2     joe@foo.com    20061213 | second\n'
                                  '1.1.1 barry@foo.com  20061213 | third\n'
                                  '1.2.1 jerry@foo.com  20061213 | fourth\n'
                                  '1.3.1 george@foo.com 20061213 | fifth\n'
                                  '                              | sixth\n',
                                  builder.get_branch(), 'a', b'rev-6',
                                  verbose=True, full=False)

        self.assertBranchAnnotate('1     joe@foo.com    20061213 | first\n'
                                  '2     joe@foo.com    20061213 | second\n'
                                  '1.1.1 barry@foo.com  20061213 | third\n'
                                  '1.2.1 jerry@foo.com  20061213 | fourth\n'
                                  '1.3.1 george@foo.com 20061213 | fifth\n'
                                  '1.3.1 george@foo.com 20061213 | sixth\n',
                                  builder.get_branch(), 'a', b'rev-6',
                                  verbose=True, full=True)

    def test_annotate_uses_branch_context(self):
        """Dotted revnos should use the Branch context.

        When annotating a non-mainline revision, the annotation should still
        use dotted revnos from the mainline.
        """
        builder = self.create_deeply_merged_trees()

        self.assertBranchAnnotate('1     joe@foo | first\n'
                                  '1.1.1 barry@f | third\n'
                                  '1.2.1 jerry@f | fourth\n'
                                  '1.3.1 george@ | fifth\n'
                                  '              | sixth\n',
                                  builder.get_branch(), 'a', b'rev-1_3_1',
                                  verbose=False, full=False)

    def test_annotate_show_ids(self):
        builder = self.create_deeply_merged_trees()

        # It looks better with real revision ids :)
        self.assertBranchAnnotate('    rev-1 | first\n'
                                  '    rev-2 | second\n'
                                  'rev-1_1_1 | third\n'
                                  'rev-1_2_1 | fourth\n'
                                  'rev-1_3_1 | fifth\n'
                                  '          | sixth\n',
                                  builder.get_branch(), 'a', b'rev-6',
                                  show_ids=True, full=False)

        self.assertBranchAnnotate('    rev-1 | first\n'
                                  '    rev-2 | second\n'
                                  'rev-1_1_1 | third\n'
                                  'rev-1_2_1 | fourth\n'
                                  'rev-1_3_1 | fifth\n'
                                  'rev-1_3_1 | sixth\n',
                                  builder.get_branch(), 'a', b'rev-6',
                                  show_ids=True, full=True)

    def test_annotate_unicode_author(self):
        tree1 = self.make_branch_and_tree('tree1')

        self.build_tree_contents([('tree1/a', b'adi\xc3\xb3s')])
        tree1.add(['a'], ids=[b'a-id'])
        tree1.commit('a', rev_id=b'rev-1',
                     committer='Pepe P\xe9rez <pperez@ejemplo.com>',
                     timestamp=1166046000.00, timezone=0)

        self.build_tree_contents([('tree1/b', b'bye')])
        tree1.add(['b'], ids=[b'b-id'])
        tree1.commit('b', rev_id=b'rev-2',
                     committer='p\xe9rez',
                     timestamp=1166046000.00, timezone=0)

        tree1.lock_read()
        self.addCleanup(tree1.unlock)

        revtree_1 = tree1.branch.repository.revision_tree(b'rev-1')
        revtree_2 = tree1.branch.repository.revision_tree(b'rev-2')

        # this passes if no exception is raised
        to_file = StringIO()
        annotate.annotate_file_tree(revtree_1, 'a',
                                    to_file=to_file, branch=tree1.branch)

        sio = BytesIO()
        to_file = codecs.getwriter('ascii')(sio, 'replace')
        annotate.annotate_file_tree(revtree_2, 'b',
                                    to_file=to_file, branch=tree1.branch)
        self.assertEqualDiff(b'2   p?rez   | bye\n', sio.getvalue())

        # test now with unicode file-like
        to_file = StringIOWithEncoding()
        annotate.annotate_file_tree(revtree_2, 'b',
                                    to_file=to_file, branch=tree1.branch)
        self.assertContainsRe('2   p\xe9rez   | bye\n', to_file.getvalue())

    def test_annotate_author_or_committer(self):
        tree1 = self.make_branch_and_tree('tree1')

        self.build_tree_contents([('tree1/a', b'hello')])
        tree1.add(['a'], ids=[b'a-id'])
        tree1.commit('a', rev_id=b'rev-1',
                     committer='Committer <committer@example.com>',
                     timestamp=1166046000.00, timezone=0)

        self.build_tree_contents([('tree1/b', b'bye')])
        tree1.add(['b'], ids=[b'b-id'])
        tree1.commit('b', rev_id=b'rev-2',
                     committer='Committer <committer@example.com>',
                     authors=['Author <author@example.com>'],
                     timestamp=1166046000.00, timezone=0)

        tree1.lock_read()
        self.addCleanup(tree1.unlock)

        self.assertBranchAnnotate('1   committ | hello\n', tree1.branch,
                                  'a', b'rev-1')

        self.assertBranchAnnotate('2   author@ | bye\n', tree1.branch,
                                  'b', b'rev-2')


class TestReannotate(tests.TestCase):

    def annotateEqual(self, expected, parents, newlines, revision_id,
                      blocks=None):
        annotate_list = list(annotate.reannotate(parents, newlines,
                                                 revision_id, blocks))
        self.assertEqual(len(expected), len(annotate_list))
        for e, a in zip(expected, annotate_list):
            self.assertEqual(e, a)

    def test_reannotate(self):
        self.annotateEqual(parent_1, [parent_1], new_1, b'blahblah')
        self.annotateEqual(expected_2_1, [parent_2], new_1, b'blahblah')
        self.annotateEqual(expected_1_2_2, [parent_1, parent_2], new_2,
                           b'blahblah')

    def test_reannotate_no_parents(self):
        self.annotateEqual(expected_1, [], new_1, b'blahblah')

    def test_reannotate_left_matching_blocks(self):
        """Ensure that left_matching_blocks has an impact.

        In this case, the annotation is ambiguous, so the hint isn't actually
        lying.
        """
        parent = [(b'rev1', b'a\n')]
        new_text = [b'a\n', b'a\n']
        blocks = [(0, 0, 1), (1, 2, 0)]
        self.annotateEqual([(b'rev1', b'a\n'), (b'rev2', b'a\n')], [parent],
                           new_text, b'rev2', blocks)
        blocks = [(0, 1, 1), (1, 2, 0)]
        self.annotateEqual([(b'rev2', b'a\n'), (b'rev1', b'a\n')], [parent],
                           new_text, b'rev2', blocks)



#    quilt.py -- Quilt patch handling
#    Copyright (C) 2011 Canonical Ltd.
#    Copyright (C) 2019 Jelmer Verooij <jelmer@jelmer.uk>
#
#    Breezy is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Breezy is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Breezy; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

"""Quilt patch handling."""

import errno
import os
import signal
import subprocess

from ... import errors, osutils, trace
from ... import transport as _mod_transport

DEFAULT_PATCHES_DIR = 'patches'
DEFAULT_SERIES_FILE = 'series'


class QuiltError(errors.BzrError):

    _fmt = "An error (%(retcode)d) occurred running quilt: %(stderr)s%(extra)s"

    def __init__(self, retcode, stdout, stderr):
        self.retcode = retcode
        self.stderr = stderr
        if stdout is not None:
            self.extra = "\n\n%s" % stdout
        else:
            self.extra = ""
        self.stdout = stdout


class QuiltNotInstalled(errors.BzrError):

    _fmt = "Quilt is not installed."


def run_quilt(
        args, working_dir, series_file=None, patches_dir=None, quiet=None):
    """Run quilt.

    :param args: Arguments to quilt
    :param working_dir: Working dir
    :param series_file: Optional path to the series file
    :param patches_dir: Optional path to the patches
    :param quilt: Whether to be quiet (quilt stderr not to terminal)
    :raise QuiltError: When running quilt fails
    """
    def subprocess_setup():
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    env = {}
    if patches_dir is not None:
        env["QUILT_PATCHES"] = patches_dir
    else:
        env["QUILT_PATCHES"] = os.path.join(working_dir, DEFAULT_PATCHES_DIR)
    if series_file is not None:
        env["QUILT_SERIES"] = series_file
    else:
        env["QUILT_SERIES"] = DEFAULT_SERIES_FILE
    # Hide output if -q is in use.
    if quiet is None:
        quiet = trace.is_quiet()
    if not quiet:
        stderr = subprocess.STDOUT
    else:
        stderr = subprocess.PIPE
    quilt_path = osutils.find_executable_on_path("quilt")
    if quilt_path is None:
        raise QuiltNotInstalled()
    command = [quilt_path] + args
    trace.mutter("running: %r", command)
    if not os.path.isdir(working_dir):
        raise AssertionError("%s is not a valid directory" % working_dir)
    try:
        proc = subprocess.Popen(
            command, cwd=working_dir, env=env,
            stdin=subprocess.PIPE, preexec_fn=subprocess_setup,
            stdout=subprocess.PIPE, stderr=stderr)
    except FileNotFoundError as e:
        raise QuiltNotInstalled() from e
    (stdout, stderr) = proc.communicate()
    if proc.returncode not in (0, 2):
        if stdout is not None:
            stdout = stdout.decode()
        if stderr is not None:
            stderr = stderr.decode()
        raise QuiltError(proc.returncode, stdout, stderr)
    if stdout is None:
        return ""
    return stdout


def quilt_pop_all(
        working_dir, patches_dir=None, series_file=None, quiet=None,
        force=False, refresh=False):
    """Pop all patches.

    :param working_dir: Directory to work in
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    args = ["pop", "-a"]
    if force:
        args.append("-f")
    if refresh:
        args.append("--refresh")
    return run_quilt(
        args, working_dir=working_dir,
        patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_pop(working_dir, patch, patches_dir=None, series_file=None, quiet=None):
    """Pop a patch.

    :param working_dir: Directory to work in
    :param patch: Patch to apply
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    return run_quilt(
        ["pop", patch], working_dir=working_dir,
        patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_push_all(working_dir, patches_dir=None, series_file=None, quiet=None,
                   force=False, refresh=False):
    """Push all patches.

    :param working_dir: Directory to work in
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    args = ["push", "-a"]
    if force:
        args.append("-f")
    if refresh:
        args.append("--refresh")
    return run_quilt(
        args, working_dir=working_dir,
        patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_push(working_dir, patch, patches_dir=None, series_file=None,
               quiet=None, force=False, refresh=False):
    """Push a patch.

    :param working_dir: Directory to work in
    :param patch: Patch to push
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    :param force: Force push
    :param refresh: Refresh
    """
    args = []
    if force:
        args.append("-f")
    if refresh:
        args.append("--refresh")
    return run_quilt(
        ["push", patch] + args, working_dir=working_dir,
        patches_dir=patches_dir, series_file=series_file, quiet=quiet)


def quilt_delete(working_dir, patch, patches_dir=None, series_file=None,
                 remove=False):
    """Delete a patch.

    :param working_dir: Directory to work in
    :param patch: Patch to push
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    :param remove: Remove the patch file as well
    """
    args = []
    if remove:
        args.append("-r")
    return run_quilt(
        ["delete", patch] + args, working_dir=working_dir,
        patches_dir=patches_dir, series_file=series_file)


def quilt_upgrade(working_dir):
    return run_quilt(["upgrade"], working_dir=working_dir)


def quilt_applied(tree):
    """Find the list of applied quilt patches.

    """
    try:
        return [os.fsdecode(patch.rstrip(b"\n"))
                for patch in tree.get_file_lines(".pc/applied-patches")
                if patch.strip() != b""]
    except _mod_transport.NoSuchFile:
        return []
    except OSError as e:
        if e.errno == errno.ENOENT:
            # File has already been removed
            return []
        raise


def quilt_unapplied(working_dir, patches_dir=None, series_file=None):
    """Find the list of unapplied quilt patches.

    :param working_dir: Directory to work in
    :param patches_dir: Optional patches directory
    :param series_file: Optional series file
    """
    working_dir = os.path.abspath(working_dir)
    if patches_dir is None:
        patches_dir = os.path.join(working_dir, DEFAULT_PATCHES_DIR)
    try:
        unapplied_patches = run_quilt(
            ["unapplied"],
            working_dir=working_dir, patches_dir=patches_dir,
            series_file=series_file).splitlines()
        patch_names = []
        for patch in unapplied_patches:
            patch = os.fsdecode(patch)
            patch_names.append(os.path.relpath(patch, patches_dir))
        return patch_names
    except QuiltError as e:
        if e.retcode == 1:
            return []
        raise


def quilt_series(tree, series_path):
    """Find the list of patches.

    :param tree: Tree to read from
    """
    try:
        return [os.fsdecode(patch.rstrip(b"\n")) for patch in
                tree.get_file_lines(series_path)
                if patch.strip() != b""]
    except OSError as e:
        if e.errno == errno.ENOENT:
            # File has already been removed
            return []
        raise
    except _mod_transport.NoSuchFile:
        return []



# Copyright (C) 2017 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from typing import TYPE_CHECKING

from .. import config, controldir, errors, pyutils, registry
from .. import transport as _mod_transport
from ..branch import format_registry as branch_format_registry
from ..repository import format_registry as repository_format_registry
from ..workingtree import format_registry as workingtree_format_registry

if TYPE_CHECKING:
    from .bzrdir import BzrDirFormat


class LineEndingError(errors.BzrError):

    _fmt = ("Line ending corrupted for file: %(file)s; "
            "Maybe your files got corrupted in transport?")

    def __init__(self, file):
        self.file = file


class BzrProber(controldir.Prober):
    """Prober for formats that use a .bzr/ control directory."""

    formats = registry.FormatRegistry["BzrDirFormat"](controldir.network_format_registry)
    """The known .bzr formats."""

    @classmethod
    def priority(klass, transport):
        return 10

    @classmethod
    def probe_transport(klass, transport):
        """Return the .bzrdir style format present in a directory."""
        try:
            format_string = transport.get_bytes(".bzr/branch-format")
        except _mod_transport.NoSuchFile:
            raise errors.NotBranchError(path=transport.base)
        except errors.BadHttpRequest as e:
            if e.reason == 'no such method: .bzr':
                # hgweb
                raise errors.NotBranchError(path=transport.base)
            raise

        try:
            first_line = format_string[:format_string.index(b"\n") + 1]
        except ValueError:
            first_line = format_string
        if (first_line.startswith(b'<!DOCTYPE') or
                first_line.startswith(b'<html')):
            raise errors.NotBranchError(
                path=transport.base, detail="format file looks like HTML")
        try:
            cls = klass.formats.get(first_line)
        except KeyError:
            if first_line.endswith(b"\r\n"):
                raise LineEndingError(file=".bzr/branch-format")
            else:
                raise errors.UnknownFormatError(
                    format=first_line, kind='bzrdir')
        return cls.from_string(format_string)

    @classmethod
    def known_formats(cls):
        result = []
        for name, format in cls.formats.items():
            if callable(format):
                format = format()
            result.append(format)
        return result


controldir.ControlDirFormat.register_prober(BzrProber)


class RemoteBzrProber(controldir.Prober):
    """Prober for remote servers that provide a Bazaar smart server."""

    @classmethod
    def priority(klass, transport):
        return -10

    @classmethod
    def probe_transport(klass, transport):
        """Return a RemoteBzrDirFormat object if it looks possible."""
        try:
            medium = transport.get_smart_medium()
        except (NotImplementedError, AttributeError,
                errors.TransportNotPossible, errors.NoSmartMedium,
                errors.SmartProtocolError):
            # no smart server, so not a branch for this format type.
            raise errors.NotBranchError(path=transport.base)
        else:
            # Decline to open it if the server doesn't support our required
            # version (3) so that the VFS-based transport will do it.
            if medium.should_probe():
                try:
                    server_version = medium.protocol_version()
                except errors.SmartProtocolError:
                    # Apparently there's no usable smart server there, even though
                    # the medium supports the smart protocol.
                    raise errors.NotBranchError(path=transport.base)
                if server_version != '2':
                    raise errors.NotBranchError(path=transport.base)
            from .remote import RemoteBzrDirFormat
            return RemoteBzrDirFormat()

    @classmethod
    def known_formats(cls):
        from .remote import RemoteBzrDirFormat
        return [RemoteBzrDirFormat()]


controldir.ControlDirFormat.register_prober(RemoteBzrProber)

# Register bzr formats
BzrProber.formats.register_lazy(
    b"Bazaar-NG meta directory, format 1\n",
    __name__ + '.bzrdir', 'BzrDirMetaFormat1')
BzrProber.formats.register_lazy(
    b"Bazaar meta directory, format 1 (with colocated branches)\n",
    __name__ + '.bzrdir', 'BzrDirMetaFormat1Colo')


def register_metadir(registry, key,
                     repository_format, help, native=True, deprecated=False,
                     branch_format=None,
                     tree_format=None,
                     hidden=False,
                     experimental=False,
                     bzrdir_format=None):
    """Register a metadir subformat.

    These all use a meta bzrdir, but can be parameterized by the
    Repository/Branch/WorkingTreeformats.

    :param repository_format: The fully-qualified repository format class
        name as a string.
    :param branch_format: Fully-qualified branch format class name as
        a string.
    :param tree_format: Fully-qualified tree format class name as
        a string.
    """
    if bzrdir_format is None:
        bzrdir_format = 'breezy.bzr.bzrdir.BzrDirMetaFormat1'
    # This should be expanded to support setting WorkingTree and Branch
    # formats, once the API supports that.

    def _load(full_name):
        mod_name, factory_name = full_name.rsplit('.', 1)
        try:
            factory = pyutils.get_named_object(mod_name, factory_name)
        except ImportError as e:
            raise ImportError('failed to load {}: {}'.format(full_name, e))
        except AttributeError:
            raise AttributeError('no factory %s in module %r'
                                 % (full_name, sys.modules[mod_name]))
        return factory()

    def helper():
        bd = _load(bzrdir_format)
        if branch_format is not None:
            bd.set_branch_format(_load(branch_format))
        if tree_format is not None:
            bd.workingtree_format = _load(tree_format)
        if repository_format is not None:
            bd.repository_format = _load(repository_format)
        return bd
    registry.register(key, helper, help, native, deprecated, hidden,
                      experimental)


# formats which have no format string are not discoverable
# and not independently creatable, so are not registered.
branch_format_registry.register_lazy(
    b"Bazaar-NG branch format 5\n", "breezy.bzr.fullhistory",
    "BzrBranchFormat5")
branch_format_registry.register_lazy(
    b"Bazaar Branch Format 6 (bzr 0.15)\n",
    "breezy.bzr.branch", "BzrBranchFormat6")
branch_format_registry.register_lazy(
    b"Bazaar Branch Format 7 (needs bzr 1.6)\n",
    "breezy.bzr.branch", "BzrBranchFormat7")
branch_format_registry.register_lazy(
    b"Bazaar Branch Format 8 (needs bzr 1.15)\n",
    "breezy.bzr.branch", "BzrBranchFormat8")
branch_format_registry.register_lazy(
    b"Bazaar-NG Branch Reference Format 1\n",
    "breezy.bzr.branch", "BranchReferenceFormat")

branch_format_registry.set_default_key(b"Bazaar Branch Format 7 (needs bzr 1.6)\n")


# formats which have no format string are not discoverable or independently
# creatable on disk, so are not registered in format_registry.  They're
# all in breezy.bzr.knitreponow.  When an instance of one of these is
# needed, it's constructed directly by the ControlDir.  Non-native formats where
# the repository is not separately opened are similar.

repository_format_registry.register_lazy(
    b'Bazaar-NG Knit Repository Format 1',
    'breezy.bzr.knitrepo',
    'RepositoryFormatKnit1',
    )

repository_format_registry.register_lazy(
    b'Bazaar Knit Repository Format 3 (bzr 0.15)\n',
    'breezy.bzr.knitrepo',
    'RepositoryFormatKnit3',
    )

repository_format_registry.register_lazy(
    b'Bazaar Knit Repository Format 4 (bzr 1.0)\n',
    'breezy.bzr.knitrepo',
    'RepositoryFormatKnit4',
    )

# Pack-based formats. There is one format for pre-subtrees, and one for
# post-subtrees to allow ease of testing.
# NOTE: These are experimental in 0.92. Stable in 1.0 and above
repository_format_registry.register_lazy(
    b'Bazaar pack repository format 1 (needs bzr 0.92)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack1',
    )
repository_format_registry.register_lazy(
    b'Bazaar pack repository format 1 with subtree support (needs bzr 0.92)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack3',
    )
repository_format_registry.register_lazy(
    b'Bazaar pack repository format 1 with rich root (needs bzr 1.0)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack4',
    )
repository_format_registry.register_lazy(
    b'Bazaar RepositoryFormatKnitPack5 (bzr 1.6)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack5',
    )
repository_format_registry.register_lazy(
    b'Bazaar RepositoryFormatKnitPack5RichRoot (bzr 1.6.1)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack5RichRoot',
    )
repository_format_registry.register_lazy(
    b'Bazaar RepositoryFormatKnitPack5RichRoot (bzr 1.6)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack5RichRootBroken',
    )
repository_format_registry.register_lazy(
    b'Bazaar RepositoryFormatKnitPack6 (bzr 1.9)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack6',
    )
repository_format_registry.register_lazy(
    b'Bazaar RepositoryFormatKnitPack6RichRoot (bzr 1.9)\n',
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatKnitPack6RichRoot',
    )
repository_format_registry.register_lazy(
    b'Bazaar repository format 2a (needs bzr 1.16 or later)\n',
    'breezy.bzr.groupcompress_repo',
    'RepositoryFormat2a',
    )

# Development formats.
# Check their docstrings to see if/when they are obsolete.
repository_format_registry.register_lazy(
    (b"Bazaar development format 2 with subtree support "
        b"(needs bzr.dev from before 1.8)\n"),
    'breezy.bzr.knitpack_repo',
    'RepositoryFormatPackDevelopment2Subtree',
    )
repository_format_registry.register_lazy(
    b'Bazaar development format 8\n',
    'breezy.bzr.groupcompress_repo',
    'RepositoryFormat2aSubtree',
    )


workingtree_format_registry.register_lazy(b"Bazaar Working Tree Format 4 (bzr 0.15)\n",
                              "breezy.bzr.workingtree_4", "WorkingTreeFormat4")
workingtree_format_registry.register_lazy(b"Bazaar Working Tree Format 5 (bzr 1.11)\n",
                              "breezy.bzr.workingtree_4", "WorkingTreeFormat5")
workingtree_format_registry.register_lazy(b"Bazaar Working Tree Format 6 (bzr 1.14)\n",
                              "breezy.bzr.workingtree_4", "WorkingTreeFormat6")
workingtree_format_registry.register_lazy(b"Bazaar-NG Working Tree format 3",
                              "breezy.bzr.workingtree_3", "WorkingTreeFormat3")
workingtree_format_registry.set_default_key(b"Bazaar Working Tree Format 6 (bzr 1.14)\n")


register_metadir(
    controldir.format_registry, 'knit',
    'breezy.bzr.knitrepo.RepositoryFormatKnit1',
    'Format using knits.  Recommended for interoperation with bzr <= 0.14.',
    branch_format='breezy.bzr.fullhistory.BzrBranchFormat5',
    tree_format='breezy.bzr.workingtree_3.WorkingTreeFormat3',
    hidden=True,
    deprecated=True)
register_metadir(
    controldir.format_registry, 'dirstate',
    'breezy.bzr.knitrepo.RepositoryFormatKnit1',
    help='Format using dirstate for working trees. '
    'Compatible with bzr 0.8 and '
    'above when accessed over the network. Introduced in bzr 0.15.',
    branch_format='breezy.bzr.fullhistory.BzrBranchFormat5',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True)
register_metadir(
    controldir.format_registry, 'dirstate-tags',
    'breezy.bzr.knitrepo.RepositoryFormatKnit1',
    help='Variant of dirstate with support for tags. '
    'Introduced in bzr 0.15.',
    branch_format='breezy.bzr.branch.BzrBranchFormat6',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True)
register_metadir(
    controldir.format_registry, 'rich-root',
    'breezy.bzr.knitrepo.RepositoryFormatKnit4',
    help='Variant of dirstate with better handling of tree roots. '
    'Introduced in bzr 1.0.',
    branch_format='breezy.bzr.branch.BzrBranchFormat6',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True)
register_metadir(
    controldir.format_registry, 'dirstate-with-subtree',
    'breezy.bzr.knitrepo.RepositoryFormatKnit3',
    help='Variant of dirstate with support for nested trees. '
    'Introduced in 0.15.',
    branch_format='breezy.bzr.branch.BzrBranchFormat6',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    experimental=True,
    deprecated=True,
    hidden=True,
    )
register_metadir(
    controldir.format_registry, 'pack-0.92',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack1',
    help='Pack-based format used in 1.x series. Introduced in 0.92. '
    'Interoperates with bzr repositories before 0.92 but cannot be '
    'read by bzr < 0.92.',
    branch_format='breezy.bzr.branch.BzrBranchFormat6',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    deprecated=True,
    hidden=True,
    )
register_metadir(
    controldir.format_registry, 'pack-0.92-subtree',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack3',
    help='Pack-based format used in 1.x series, with subtree support. '
    'Introduced in 0.92. Interoperates with '
    'bzr repositories before 0.92 but cannot be read by bzr < 0.92.',
    branch_format='breezy.bzr.branch.BzrBranchFormat6',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True,
    experimental=True,
    )
register_metadir(
    controldir.format_registry, 'rich-root-pack',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack4',
    help='A variant of pack-0.92 that supports rich-root data '
    '(needed for bzr-svn and bzr-git). Introduced in 1.0.',
    branch_format='breezy.bzr.branch.BzrBranchFormat6',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True,
    )
register_metadir(
    controldir.format_registry, '1.6',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack5',
    help='A format that allows a branch to indicate that there is another '
    '(stacked) repository that should be used to access data that is '
    'not present locally.',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True,
    )
register_metadir(
    controldir.format_registry, '1.6.1-rich-root',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack5RichRoot',
    help='A variant of 1.6 that supports rich-root data '
    '(needed for bzr-svn and bzr-git).',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True,
    )
register_metadir(
    controldir.format_registry, '1.9',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6',
    help='A repository format using B+tree indexes. These indexes '
    'are smaller in size, have smarter caching and provide faster '
    'performance for most operations.',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True,
    )
register_metadir(
    controldir.format_registry, '1.9-rich-root',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6RichRoot',
    help='A variant of 1.9 that supports rich-root data '
    '(needed for bzr-svn and bzr-git).',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat4',
    hidden=True,
    deprecated=True,
    )
register_metadir(
    controldir.format_registry, '1.14',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6',
    help='A working-tree format that supports content filtering.',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat5',
    hidden=True,
    deprecated=True,
    )
register_metadir(
    controldir.format_registry, '1.14-rich-root',
    'breezy.bzr.knitpack_repo.RepositoryFormatKnitPack6RichRoot',
    help='A variant of 1.14 that supports rich-root data '
    '(needed for bzr-svn and bzr-git).',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat5',
    hidden=True,
    deprecated=True,
    )
# The following un-numbered 'development' formats should always just be aliases.
register_metadir(
    controldir.format_registry, 'development-subtree',
    'breezy.bzr.groupcompress_repo.RepositoryFormat2aSubtree',
    help='Current development format, subtree variant. Can convert data to and '
         'from pack-0.92-subtree (and anything compatible with '
         'pack-0.92-subtree) format repositories. Repositories and branches in '
         'this format can only be read by bzr.dev. Please read '
         'https://www.breezy-vcs.org/developers/development-repo.html '
         'before use.',
    branch_format='breezy.bzr.branch.BzrBranchFormat8',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat6',
    experimental=True,
    deprecated=True,
    hidden=True,
    )
register_metadir(
    controldir.format_registry, 'development5-subtree',
    'breezy.bzr.knitpack_repo.RepositoryFormatPackDevelopment2Subtree',
    help='Development format, subtree variant. Can convert data to and '
    'from pack-0.92-subtree (and anything compatible with '
    'pack-0.92-subtree) format repositories. Repositories and branches in '
    'this format can only be read by bzr.dev. Please read '
    'https://www.breezy-vcs.org/developers/development-repo.html '
    'before use.',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat6',
    experimental=True,
    deprecated=True,
    hidden=True,
    )

register_metadir(
    controldir.format_registry, 'development-colo',
    'breezy.bzr.groupcompress_repo.RepositoryFormat2a',
    help='The 2a format with experimental support for colocated branches.\n',
    branch_format='breezy.bzr.branch.BzrBranchFormat7',
    tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat6',
    experimental=True,
    bzrdir_format='breezy.bzr.bzrdir.BzrDirMetaFormat1Colo',
    hidden=True,
    )


# And the development formats above will have aliased one of the following:

# Finally, the current format.
register_metadir(controldir.format_registry, '2a',
                 'breezy.bzr.groupcompress_repo.RepositoryFormat2a',
                 help='Format for the bzr 2.0 series.\n',
                 branch_format='breezy.bzr.branch.BzrBranchFormat7',
                 tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat6',
                 experimental=False,
                 deprecated=False,
                 )

# The following format should be an alias for the rich root equivalent
# of the default format

controldir.format_registry.register_alias(
    'default-rich-root', '2a', hidden=True)

# The following format should is just an alias for the default bzr format.
controldir.format_registry.register_alias('bzr', '2a')

# The current format that is made on 'bzr init'.
format_name = config.GlobalStack().get('default_format')
controldir.format_registry.set_default(format_name)



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from .. import bugtracker, urlutils
from . import TestCase, TestCaseWithMemoryTransport


class ErrorsTest(TestCaseWithMemoryTransport):

    def test_unknown_bug_tracker_abbreviation(self):
        """Test the formatting of UnknownBugTrackerAbbreviation."""
        branch = self.make_branch('some_branch')
        error = bugtracker.UnknownBugTrackerAbbreviation('xxx', branch)
        self.assertEqual(
            "Cannot find registered bug tracker called xxx on %s" % branch,
            str(error))

    def test_malformed_bug_identifier(self):
        """Test the formatting of MalformedBugIdentifier."""
        error = bugtracker.MalformedBugIdentifier(
            'bogus', 'reason for bogosity')
        self.assertEqual(
            'Did not understand bug identifier bogus: reason for bogosity. '
            'See "brz help bugs" for more information on this feature.',
            str(error))

    def test_incorrect_url(self):
        err = bugtracker.InvalidBugTrackerURL('foo', 'http://bug.example.com/')
        self.assertEqual(
            ("The URL for bug tracker \"foo\" doesn't contain {id}: "
             "http://bug.example.com/"),
            str(err))


class TestGetBugURL(TestCaseWithMemoryTransport):
    """Tests for bugtracker.get_bug_url"""

    class TransientTracker:
        """An transient tracker used for testing."""

        @classmethod
        def get(klass, abbreviation, branch):
            klass.log.append(('get', abbreviation, branch))
            if abbreviation != 'transient':
                return None
            return klass()

        def get_bug_url(self, bug_id):
            self.log.append(('get_bug_url', bug_id))
            return "http://bugs.example.com/%s" % bug_id

    def setUp(self):
        super().setUp()
        self.tracker_type = TestGetBugURL.TransientTracker
        self.tracker_type.log = []
        bugtracker.tracker_registry.register('transient', self.tracker_type)
        self.addCleanup(bugtracker.tracker_registry.remove, 'transient')

    def test_get_bug_url_for_transient_tracker(self):
        branch = self.make_branch('some_branch')
        self.assertEqual('http://bugs.example.com/1234',
                         bugtracker.get_bug_url('transient', branch, '1234'))
        self.assertEqual(
            [('get', 'transient', branch), ('get_bug_url', '1234')],
            self.tracker_type.log)

    def test_unrecognized_abbreviation_raises_error(self):
        """If the abbreviation is unrecognized, then raise an error."""
        branch = self.make_branch('some_branch')
        self.assertRaises(bugtracker.UnknownBugTrackerAbbreviation,
                          bugtracker.get_bug_url, 'xxx', branch, '1234')
        self.assertEqual([('get', 'xxx', branch)], self.tracker_type.log)


class TestBuiltinTrackers(TestCaseWithMemoryTransport):
    """Test that the builtin trackers are registered and return sane URLs."""

    def test_launchpad_registered(self):
        """The Launchpad bug tracker should be registered by default and
        generate Launchpad bug page URLs.
        """
        branch = self.make_branch('some_branch')
        tracker = bugtracker.tracker_registry.get_tracker('lp', branch)
        self.assertEqual('https://launchpad.net/bugs/1234',
                         tracker.get_bug_url('1234'))

    def test_debian_registered(self):
        """The Debian bug tracker should be registered by default and generate
        bugs.debian.org bug page URLs.
        """
        branch = self.make_branch('some_branch')
        tracker = bugtracker.tracker_registry.get_tracker('deb', branch)
        self.assertEqual('http://bugs.debian.org/1234',
                         tracker.get_bug_url('1234'))

    def test_gnome_registered(self):
        branch = self.make_branch('some_branch')
        tracker = bugtracker.tracker_registry.get_tracker('gnome', branch)
        self.assertEqual('http://bugzilla.gnome.org/show_bug.cgi?id=1234',
                         tracker.get_bug_url('1234'))

    def test_trac_registered(self):
        """The Trac bug tracker should be registered by default and generate
        Trac bug page URLs when the appropriate configuration is present.
        """
        branch = self.make_branch('some_branch')
        config = branch.get_config()
        config.set_user_option('trac_foo_url', 'http://bugs.example.com/trac')
        tracker = bugtracker.tracker_registry.get_tracker('foo', branch)
        self.assertEqual('http://bugs.example.com/trac/ticket/1234',
                         tracker.get_bug_url('1234'))

    def test_bugzilla_registered(self):
        """The Bugzilla bug tracker should be registered by default and
        generate Bugzilla bug page URLs when the appropriate configuration is
        present.
        """
        branch = self.make_branch('some_branch')
        config = branch.get_config()
        config.set_user_option('bugzilla_foo_url', 'http://bugs.example.com')
        tracker = bugtracker.tracker_registry.get_tracker('foo', branch)
        self.assertEqual('http://bugs.example.com/show_bug.cgi?id=1234',
                         tracker.get_bug_url('1234'))

    def test_github(self):
        branch = self.make_branch('some_branch')
        tracker = bugtracker.tracker_registry.get_tracker('github', branch)
        self.assertEqual('https://github.com/breezy-team/breezy/issues/1234',
                         tracker.get_bug_url('breezy-team/breezy/1234'))

    def test_generic_registered(self):
        branch = self.make_branch('some_branch')
        config = branch.get_config()
        config.set_user_option('bugtracker_foo_url',
                               'http://bugs.example.com/{id}/view.html')
        tracker = bugtracker.tracker_registry.get_tracker('foo', branch)
        self.assertEqual('http://bugs.example.com/1234/view.html',
                         tracker.get_bug_url('1234'))

    def test_generic_registered_non_integer(self):
        branch = self.make_branch('some_branch')
        config = branch.get_config()
        config.set_user_option('bugtracker_foo_url',
                               'http://bugs.example.com/{id}/view.html')
        tracker = bugtracker.tracker_registry.get_tracker('foo', branch)
        self.assertEqual('http://bugs.example.com/ABC-1234/view.html',
                         tracker.get_bug_url('ABC-1234'))

    def test_generic_incorrect_url(self):
        branch = self.make_branch('some_branch')
        config = branch.get_config()
        config.set_user_option('bugtracker_foo_url',
                               'http://bugs.example.com/view.html')
        tracker = bugtracker.tracker_registry.get_tracker('foo', branch)
        self.assertRaises(bugtracker.InvalidBugTrackerURL, tracker.get_bug_url,
                          '1234')


class TestUniqueIntegerBugTracker(TestCaseWithMemoryTransport):

    def test_appends_id_to_base_url(self):
        """The URL of a bug is the base URL joined to the identifier."""
        tracker = bugtracker.UniqueIntegerBugTracker('xxx',
                                                     'http://bugs.example.com/foo')
        self.assertEqual('http://bugs.example.com/foo1234',
                         tracker.get_bug_url('1234'))

    def test_returns_tracker_if_abbreviation_matches(self):
        """The get() method should return an instance of the tracker if the
        given abbreviation matches the tracker's abbreviated name.
        """
        tracker = bugtracker.UniqueIntegerBugTracker('xxx',
                                                     'http://bugs.example.com/')
        branch = self.make_branch('some_branch')
        self.assertIs(tracker, tracker.get('xxx', branch))

    def test_returns_none_if_abbreviation_doesnt_match(self):
        """The get() method should return None if the given abbreviated name
        doesn't match the tracker's abbreviation.
        """
        tracker = bugtracker.UniqueIntegerBugTracker('xxx',
                                                     'http://bugs.example.com/')
        branch = self.make_branch('some_branch')
        self.assertIs(None, tracker.get('yyy', branch))

    def test_doesnt_consult_branch(self):
        """A UniqueIntegerBugTracker shouldn't consult the branch for tracker
        information.
        """
        tracker = bugtracker.UniqueIntegerBugTracker('xxx',
                                                     'http://bugs.example.com/')
        self.assertIs(tracker, tracker.get('xxx', None))
        self.assertIs(None, tracker.get('yyy', None))

    def test_check_bug_id_only_accepts_integers(self):
        """A UniqueIntegerBugTracker accepts integers as bug IDs."""
        tracker = bugtracker.UniqueIntegerBugTracker('xxx',
                                                     'http://bugs.example.com/')
        tracker.check_bug_id('1234')

    def test_check_bug_id_doesnt_accept_non_integers(self):
        """A UniqueIntegerBugTracker rejects non-integers as bug IDs."""
        tracker = bugtracker.UniqueIntegerBugTracker('xxx',
                                                     'http://bugs.example.com/')
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, tracker.check_bug_id, 'red')


class TestProjectIntegerBugTracker(TestCaseWithMemoryTransport):

    def test_appends_id_to_base_url(self):
        """The URL of a bug is the base URL joined to the identifier."""
        tracker = bugtracker.ProjectIntegerBugTracker('xxx',
                                                      'http://bugs.example.com/{project}/{id}')
        self.assertEqual('http://bugs.example.com/foo/1234',
                         tracker.get_bug_url('foo/1234'))

    def test_returns_tracker_if_abbreviation_matches(self):
        """The get() method should return an instance of the tracker if the
        given abbreviation matches the tracker's abbreviated name.
        """
        tracker = bugtracker.ProjectIntegerBugTracker('xxx',
                                                      'http://bugs.example.com/{project}/{id}')
        branch = self.make_branch('some_branch')
        self.assertIs(tracker, tracker.get('xxx', branch))

    def test_returns_none_if_abbreviation_doesnt_match(self):
        """The get() method should return None if the given abbreviated name
        doesn't match the tracker's abbreviation.
        """
        tracker = bugtracker.ProjectIntegerBugTracker('xxx',
                                                      'http://bugs.example.com/{project}/{id}')
        branch = self.make_branch('some_branch')
        self.assertIs(None, tracker.get('yyy', branch))

    def test_doesnt_consult_branch(self):
        """Shouldn't consult the branch for tracker information.
        """
        tracker = bugtracker.ProjectIntegerBugTracker('xxx',
                                                      'http://bugs.example.com/{project}/{id}')
        self.assertIs(tracker, tracker.get('xxx', None))
        self.assertIs(None, tracker.get('yyy', None))

    def test_check_bug_id_only_accepts_project_integers(self):
        """Accepts integers as bug IDs."""
        tracker = bugtracker.ProjectIntegerBugTracker('xxx',
                                                      'http://bugs.example.com/{project}/{id}')
        tracker.check_bug_id('project/1234')

    def test_check_bug_id_doesnt_accept_non_project_integers(self):
        """Rejects non-integers as bug IDs."""
        tracker = bugtracker.ProjectIntegerBugTracker('xxx',
                                                      'http://bugs.example.com/{project}/{id}')
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, tracker.check_bug_id, 'red')
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, tracker.check_bug_id, '1234')


class TestURLParametrizedBugTracker(TestCaseWithMemoryTransport):
    """Tests for URLParametrizedBugTracker."""

    def setUp(self):
        super().setUp()
        self.url = 'http://twistedmatrix.com/trac'
        self.tracker = bugtracker.URLParametrizedBugTracker('some', 'ticket/')

    def test_get_with_unsupported_tag(self):
        """If asked for an unrecognized or unconfigured tag, return None."""
        branch = self.make_branch('some_branch')
        self.assertEqual(None, self.tracker.get('lp', branch))
        self.assertEqual(None, self.tracker.get('twisted', branch))

    def test_get_with_supported_tag(self):
        """If asked for a valid tag, return a tracker instance that can map bug
        IDs to <base_url>/<bug_area> + <bug_id>."""
        bugtracker.tracker_registry.register('some', self.tracker)
        self.addCleanup(bugtracker.tracker_registry.remove, 'some')

        branch = self.make_branch('some_branch')
        config = branch.get_config()
        config.set_user_option('some_twisted_url', self.url)
        tracker = self.tracker.get('twisted', branch)
        self.assertEqual(
            urlutils.join(self.url, 'ticket/') + '1234',
            tracker.get_bug_url('1234'))

    def test_get_bug_url_for_integer_id(self):
        self.tracker.check_bug_id('1234')

    def test_get_bug_url_for_non_integer_id(self):
        self.tracker.check_bug_id('ABC-1234')


class TestURLParametrizedIntegerBugTracker(TestCaseWithMemoryTransport):
    """Tests for URLParametrizedIntegerBugTracker."""

    def setUp(self):
        super().setUp()
        self.url = 'http://twistedmatrix.com/trac'
        self.tracker = bugtracker.URLParametrizedIntegerBugTracker('some',
                                                                   'ticket/')

    def test_get_bug_url_for_bad_bug(self):
        """When given a bug identifier that is invalid for Trac, get_bug_url
        should raise an error.
        """
        self.assertRaises(
            bugtracker.MalformedBugIdentifier, self.tracker.get_bug_url, 'bad')


class TestPropertyEncoding(TestCase):
    """Tests for how the bug URLs are encoded as revision properties."""

    def test_encoding_one(self):
        self.assertEqual(
            'http://example.com/bugs/1 fixed',
            bugtracker.encode_fixes_bug_urls(
                [('http://example.com/bugs/1', 'fixed')]))

    def test_encoding_zero(self):
        self.assertEqual('', bugtracker.encode_fixes_bug_urls([]))

    def test_encoding_two(self):
        self.assertEqual(
            'http://example.com/bugs/1 fixed\n'
            'http://example.com/bugs/2 related',
            bugtracker.encode_fixes_bug_urls(
                [('http://example.com/bugs/1', 'fixed'),
                 ('http://example.com/bugs/2', 'related')]))

    def test_encoding_with_space(self):
        self.assertRaises(
            bugtracker.InvalidBugUrl,
            bugtracker.encode_fixes_bug_urls,
            [('http://example.com/bugs/ 1', 'fixed')])


class TestPropertyDecoding(TestCase):
    """Tests for parsing bug revision properties."""

    def test_decoding_one(self):
        self.assertEqual(
            [('http://example.com/bugs/1', 'fixed')],
            list(bugtracker.decode_bug_urls(
                'http://example.com/bugs/1 fixed')))

    def test_decoding_zero(self):
        self.assertEqual([], list(bugtracker.decode_bug_urls('')))

    def test_decoding_two(self):
        self.assertEqual(
            [('http://example.com/bugs/1', 'fixed'),
             ('http://example.com/bugs/2', 'related')],
            list(bugtracker.decode_bug_urls(
                'http://example.com/bugs/1 fixed\n'
                'http://example.com/bugs/2 related')))

    def test_decoding_invalid(self):
        self.assertRaises(
            bugtracker.InvalidLineInBugsProperty,
            list,
            bugtracker.decode_bug_urls('http://example.com/bugs/ 1 fixed\n'))



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import contextlib
import shutil
import sys
import tempfile
from io import BytesIO

import patiencediff

from . import (builtins, delta, diff, errors, osutils, patches, shelf,
               textfile, trace, ui, workingtree)
from .i18n import gettext


class UseEditor(Exception):
    """Use an editor instead of selecting hunks."""


class ShelfReporter:

    vocab = {'add file': gettext('Shelve adding file "%(path)s"?'),
             'binary': gettext('Shelve binary changes?'),
             'change kind': gettext('Shelve changing "%s" from %(other)s'
                                    ' to %(this)s?'),
             'delete file': gettext('Shelve removing file "%(path)s"?'),
             'final': gettext('Shelve %d change(s)?'),
             'hunk': gettext('Shelve?'),
             'modify target': gettext('Shelve changing target of'
                                      ' "%(path)s" from "%(other)s" to "%(this)s"?'),
             'rename': gettext('Shelve renaming "%(other)s" =>'
                               ' "%(this)s"?')
             }

    invert_diff = False

    def __init__(self):
        self.delta_reporter = delta._ChangeReporter()

    def no_changes(self):
        """Report that no changes were selected to apply."""
        trace.warning('No changes to shelve.')

    def shelved_id(self, shelf_id):
        """Report the id changes were shelved to."""
        trace.note(gettext('Changes shelved with id "%d".') % shelf_id)

    def changes_destroyed(self):
        """Report that changes were made without shelving."""
        trace.note(gettext('Selected changes destroyed.'))

    def selected_changes(self, transform):
        """Report the changes that were selected."""
        trace.note(gettext("Selected changes:"))
        changes = transform.iter_changes()
        delta.report_changes(changes, self.delta_reporter)

    def prompt_change(self, change):
        """Determine the prompt for a change to apply."""
        if change[0] == 'rename':
            vals = {'this': change[3], 'other': change[2]}
        elif change[0] == 'change kind':
            vals = {'path': change[4], 'other': change[2], 'this': change[3]}
        elif change[0] == 'modify target':
            vals = {'path': change[2], 'other': change[3], 'this': change[4]}
        else:
            vals = {'path': change[3]}
        prompt = self.vocab[change[0]] % vals
        return prompt


class ApplyReporter(ShelfReporter):

    vocab = {'add file': gettext('Delete file "%(path)s"?'),
             'binary': gettext('Apply binary changes?'),
             'change kind': gettext('Change "%(path)s" from %(this)s'
                                    ' to %(other)s?'),
             'delete file': gettext('Add file "%(path)s"?'),
             'final': gettext('Apply %d change(s)?'),
             'hunk': gettext('Apply change?'),
             'modify target': gettext('Change target of'
                                      ' "%(path)s" from "%(this)s" to "%(other)s"?'),
             'rename': gettext('Rename "%(this)s" => "%(other)s"?'),
             }

    invert_diff = True

    def changes_destroyed(self):
        pass


class Shelver:
    """Interactively shelve the changes in a working tree."""

    def __init__(self, work_tree, target_tree, diff_writer=None, auto=False,
                 auto_apply=False, file_list=None, message=None,
                 destroy=False, manager=None, reporter=None):
        """Constructor.

        :param work_tree: The working tree to shelve changes from.
        :param target_tree: The "unchanged" / old tree to compare the
            work_tree to.
        :param auto: If True, shelve each possible change.
        :param auto_apply: If True, shelve changes with no final prompt.
        :param file_list: If supplied, only files in this list may be shelved.
        :param message: The message to associate with the shelved changes.
        :param destroy: Change the working tree without storing the shelved
            changes.
        :param manager: The shelf manager to use.
        :param reporter: Object for reporting changes to user.
        """
        self.work_tree = work_tree
        self.target_tree = target_tree
        self.diff_writer = diff_writer
        if self.diff_writer is None:
            self.diff_writer = sys.stdout
        if manager is None:
            manager = work_tree.get_shelf_manager()
        self.manager = manager
        self.auto = auto
        self.auto_apply = auto_apply
        self.file_list = file_list
        self.message = message
        self.destroy = destroy
        if reporter is None:
            reporter = ShelfReporter()
        self.reporter = reporter
        config = self.work_tree.branch.get_config()
        self.change_editor = config.get_change_editor(target_tree, work_tree)
        self.work_tree.lock_tree_write()

    @classmethod
    def from_args(klass, diff_writer, revision=None, all=False, file_list=None,
                  message=None, directory=None, destroy=False):
        """Create a shelver from commandline arguments.

        The returned shelver wil have a work_tree that is locked and should
        be unlocked.

        :param revision: RevisionSpec of the revision to compare to.
        :param all: If True, shelve all changes without prompting.
        :param file_list: If supplied, only files in this list may be  shelved.
        :param message: The message to associate with the shelved changes.
        :param directory: The directory containing the working tree.
        :param destroy: Change the working tree without storing the shelved
            changes.
        """
        if directory is None:
            directory = '.'
        elif file_list:
            file_list = [osutils.pathjoin(directory, f) for f in file_list]
        tree, path = workingtree.WorkingTree.open_containing(directory)
        # Ensure that tree is locked for the lifetime of target_tree, as
        # target tree may be reading from the same dirstate.
        with tree.lock_tree_write():
            target_tree = builtins._get_one_revision_tree('shelf2', revision,
                                                          tree.branch, tree)
            files = tree.safe_relpath_files(file_list)
            return klass(tree, target_tree, diff_writer, all, all, files,
                         message, destroy)

    def run(self):
        """Interactively shelve the changes."""
        creator = shelf.ShelfCreator(self.work_tree, self.target_tree,
                                     self.file_list)
        self.tempdir = tempfile.mkdtemp()
        changes_shelved = 0
        try:
            for change in creator.iter_shelvable():
                if change[0] == 'modify text':
                    try:
                        changes_shelved += self.handle_modify_text(creator,
                                                                   change[1])
                    except errors.BinaryFile:
                        if self.prompt_bool(self.reporter.vocab['binary']):
                            changes_shelved += 1
                            creator.shelve_content_change(change[1])
                else:
                    if self.prompt_bool(self.reporter.prompt_change(change)):
                        creator.shelve_change(change)
                        changes_shelved += 1
            if changes_shelved > 0:
                self.reporter.selected_changes(creator.work_transform)
                if (self.auto_apply or self.prompt_bool(
                        self.reporter.vocab['final'] % changes_shelved)):
                    if self.destroy:
                        creator.transform()
                        self.reporter.changes_destroyed()
                    else:
                        shelf_id = self.manager.shelve_changes(creator,
                                                               self.message)
                        self.reporter.shelved_id(shelf_id)
            else:
                self.reporter.no_changes()
        finally:
            shutil.rmtree(self.tempdir)
            creator.finalize()

    def finalize(self):
        if self.change_editor is not None:
            self.change_editor.finish()
        self.work_tree.unlock()

    def get_parsed_patch(self, file_id, invert=False):
        """Return a parsed version of a file's patch.

        :param file_id: The id of the file to generate a patch for.
        :param invert: If True, provide an inverted patch (insertions displayed
            as removals, removals displayed as insertions).
        :return: A patches.Patch.
        """
        diff_file = BytesIO()
        if invert:
            old_tree = self.work_tree
            new_tree = self.target_tree
        else:
            old_tree = self.target_tree
            new_tree = self.work_tree
        old_path = old_tree.id2path(file_id)
        new_path = new_tree.id2path(file_id)
        path_encoding = osutils.get_terminal_encoding()
        text_differ = diff.DiffText(old_tree, new_tree, diff_file,
                                    path_encoding=path_encoding)
        patch = text_differ.diff(old_path, new_path, 'file', 'file')
        diff_file.seek(0)
        return patches.parse_patch(diff_file)

    def prompt(self, message, choices, default):
        return ui.ui_factory.choose(message, choices, default=default)

    def prompt_bool(self, question, allow_editor=False):
        """Prompt the user with a yes/no question.

        This may be overridden by self.auto.  It may also *set* self.auto.  It
        may also raise UserAbort.
        :param question: The question to ask the user.
        :return: True or False
        """
        if self.auto:
            return True
        alternatives_chars = 'yn'
        alternatives = '&yes\n&No'
        if allow_editor:
            alternatives_chars += 'e'
            alternatives += '\n&edit manually'
        alternatives_chars += 'fq'
        alternatives += '\n&finish\n&quit'
        choice = self.prompt(question, alternatives, 1)
        if choice is None:
            # EOF.
            char = 'n'
        else:
            char = alternatives_chars[choice]
        if char == 'y':
            return True
        elif char == 'e' and allow_editor:
            raise UseEditor
        elif char == 'f':
            self.auto = True
            return True
        if char == 'q':
            raise errors.UserAbort()
        else:
            return False

    def handle_modify_text(self, creator, file_id):
        """Handle modified text, by using hunk selection or file editing.

        :param creator: A ShelfCreator.
        :param file_id: The id of the file that was modified.
        :return: The number of changes.
        """
        path = self.work_tree.id2path(file_id)
        work_tree_lines = self.work_tree.get_file_lines(path, file_id)
        try:
            lines, change_count = self._select_hunks(creator, file_id,
                                                     work_tree_lines)
        except UseEditor:
            lines, change_count = self._edit_file(file_id, work_tree_lines)
        if change_count != 0:
            creator.shelve_lines(file_id, lines)
        return change_count

    def _select_hunks(self, creator, file_id, work_tree_lines):
        """Provide diff hunk selection for modified text.

        If self.reporter.invert_diff is True, the diff is inverted so that
        insertions are displayed as removals and vice versa.

        :param creator: a ShelfCreator
        :param file_id: The id of the file to shelve.
        :param work_tree_lines: Line contents of the file in the working tree.
        :return: number of shelved hunks.
        """
        if self.reporter.invert_diff:
            target_lines = work_tree_lines
        else:
            path = self.target_tree.id2path(file_id)
            target_lines = self.target_tree.get_file_lines(path)
        textfile.check_text_lines(work_tree_lines)
        textfile.check_text_lines(target_lines)
        parsed = self.get_parsed_patch(file_id, self.reporter.invert_diff)
        final_hunks = []
        if not self.auto:
            offset = 0
            self.diff_writer.write(parsed.get_header())
            for hunk in parsed.hunks:
                self.diff_writer.write(hunk.as_bytes())
                selected = self.prompt_bool(self.reporter.vocab['hunk'],
                                            allow_editor=(self.change_editor
                                                          is not None))
                if not self.reporter.invert_diff:
                    selected = (not selected)
                if selected:
                    hunk.mod_pos += offset
                    final_hunks.append(hunk)
                else:
                    offset -= (hunk.mod_range - hunk.orig_range)
        sys.stdout.flush()
        if self.reporter.invert_diff:
            change_count = len(final_hunks)
        else:
            change_count = len(parsed.hunks) - len(final_hunks)
        patched = patches.iter_patched_from_hunks(target_lines,
                                                  final_hunks)
        lines = list(patched)
        return lines, change_count

    def _edit_file(self, file_id, work_tree_lines):
        """
        :param file_id: id of the file to edit.
        :param work_tree_lines: Line contents of the file in the working tree.
        :return: (lines, change_region_count), where lines is the new line
            content of the file, and change_region_count is the number of
            changed regions.
        """
        lines = osutils.split_lines(self.change_editor.edit_file(
            self.change_editor.old_tree.id2path(file_id),
            self.change_editor.new_tree.id2path(file_id)))
        return lines, self._count_changed_regions(work_tree_lines, lines)

    @staticmethod
    def _count_changed_regions(old_lines, new_lines):
        matcher = patiencediff.PatienceSequenceMatcher(None, old_lines,
                                                       new_lines)
        blocks = matcher.get_matching_blocks()
        return len(blocks) - 2


class Unshelver:
    """Unshelve changes into a working tree."""

    @classmethod
    def from_args(klass, shelf_id=None, action='apply', directory='.',
                  write_diff_to=None):
        """Create an unshelver from commandline arguments.

        The returned shelver will have a tree that is locked and should
        be unlocked.

        :param shelf_id: Integer id of the shelf, as a string.
        :param action: action to perform.  May be 'apply', 'dry-run',
            'delete', 'preview'.
        :param directory: The directory to unshelve changes into.
        :param write_diff_to: See Unshelver.__init__().
        """
        tree, path = workingtree.WorkingTree.open_containing(directory)
        tree.lock_tree_write()
        try:
            manager = tree.get_shelf_manager()
            if shelf_id is not None:
                try:
                    shelf_id = int(shelf_id)
                except ValueError:
                    raise shelf.InvalidShelfId(shelf_id)
            else:
                shelf_id = manager.last_shelf()
                if shelf_id is None:
                    raise errors.CommandError(
                        gettext('No changes are shelved.'))
            apply_changes = True
            delete_shelf = True
            read_shelf = True
            show_diff = False
            if action == 'dry-run':
                apply_changes = False
                delete_shelf = False
            elif action == 'preview':
                apply_changes = False
                delete_shelf = False
                show_diff = True
            elif action == 'delete-only':
                apply_changes = False
                read_shelf = False
            elif action == 'keep':
                apply_changes = True
                delete_shelf = False
        except:
            tree.unlock()
            raise
        return klass(tree, manager, shelf_id, apply_changes, delete_shelf,
                     read_shelf, show_diff, write_diff_to)

    def __init__(self, tree, manager, shelf_id, apply_changes=True,
                 delete_shelf=True, read_shelf=True, show_diff=False,
                 write_diff_to=None):
        """Constructor.

        :param tree: The working tree to unshelve into.
        :param manager: The ShelveManager containing the shelved changes.
        :param shelf_id:
        :param apply_changes: If True, apply the shelved changes to the
            working tree.
        :param delete_shelf: If True, delete the changes from the shelf.
        :param read_shelf: If True, read the changes from the shelf.
        :param show_diff: If True, show the diff that would result from
            unshelving the changes.
        :param write_diff_to: A file-like object where the diff will be
            written to. If None, ui.ui_factory.make_output_stream() will
            be used.
        """
        self.tree = tree
        manager = tree.get_shelf_manager()
        self.manager = manager
        self.shelf_id = shelf_id
        self.apply_changes = apply_changes
        self.delete_shelf = delete_shelf
        self.read_shelf = read_shelf
        self.show_diff = show_diff
        self.write_diff_to = write_diff_to

    def run(self):
        """Perform the unshelving operation."""
        with contextlib.ExitStack() as exit_stack:
            exit_stack.enter_context(self.tree.lock_tree_write())
            if self.read_shelf:
                trace.note(gettext('Using changes with id "%d".') %
                           self.shelf_id)
                unshelver = self.manager.get_unshelver(self.shelf_id)
                exit_stack.callback(unshelver.finalize)
                if unshelver.message is not None:
                    trace.note(gettext('Message: %s') % unshelver.message)
                change_reporter = delta._ChangeReporter()
                merger = unshelver.make_merger()
                merger.change_reporter = change_reporter
                if self.apply_changes:
                    merger.do_merge()
                elif self.show_diff:
                    self.write_diff(merger)
                else:
                    self.show_changes(merger)
            if self.delete_shelf:
                self.manager.delete_shelf(self.shelf_id)
                trace.note(gettext('Deleted changes with id "%d".') %
                           self.shelf_id)

    def write_diff(self, merger):
        """Write this operation's diff to self.write_diff_to."""
        tree_merger = merger.make_merger()
        tt = tree_merger.make_preview_transform()
        new_tree = tt.get_preview_tree()
        if self.write_diff_to is None:
            self.write_diff_to = ui.ui_factory.make_output_stream(
                encoding_type='exact')
        path_encoding = osutils.get_diff_header_encoding()
        diff.show_diff_trees(merger.this_tree, new_tree, self.write_diff_to,
                             path_encoding=path_encoding)
        tt.finalize()

    def show_changes(self, merger):
        """Show the changes that this operation specifies."""
        tree_merger = merger.make_merger()
        # This implicitly shows the changes via the reporter, so we're done...
        tt = tree_merger.make_preview_transform()
        tt.finalize()



# Copyright (C) 2004, 2005 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# TODO: Perhaps rather than mapping options and arguments back and
# forth, we should just pass in the whole argv, and allow
# ExternalCommands to handle it differently to internal commands?


import os

from .commands import Command


class ExternalCommand(Command):
    """Class to wrap external commands."""

    @classmethod
    def find_command(cls, cmd):
        import os.path
        bzrpath = os.environ.get('BZRPATH', '')

        for dir in bzrpath.split(os.pathsep):
            # Empty directories are not real paths
            if not dir:
                continue
            # This needs to be os.path.join() or windows cannot
            # find the batch file that you are wanting to execute
            path = os.path.join(dir, cmd)
            if os.path.isfile(path):
                return ExternalCommand(path)

        return None

    def __init__(self, path):
        self.path = path

    def name(self):
        return os.path.basename(self.path)

    def run(self, *args, **kwargs):
        raise NotImplementedError('should not be called on %r' % self)

    def run_argv_aliases(self, argv, alias_argv=None):
        return os.spawnv(os.P_WAIT, self.path, [self.path] + argv)

    def help(self):
        m = 'external command from %s\n\n' % self.path
        pipe = os.popen('%s --help' % self.path)
        return m + pipe.read()



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Repository.add_fallback_repository."""

from breezy import errors
from breezy.revision import NULL_REVISION
from breezy.tests import TestNotApplicable
from breezy.tests.per_repository import TestCaseWithRepository


class TestAddFallbackRepository(TestCaseWithRepository):

    def test_add_fallback_repository(self):
        repo = self.make_repository('repo')
        tree = self.make_branch_and_tree('branch')
        if not repo._format.supports_external_lookups:
            self.assertRaises(errors.UnstackableRepositoryFormat,
                              repo.add_fallback_repository, tree.branch.repository)
            raise TestNotApplicable
        repo.add_fallback_repository(tree.branch.repository)
        # the repository has been added correctly if we can query against it.
        revision_id = tree.commit('1st post')
        repo.lock_read()
        self.addCleanup(repo.unlock)
        # can see all revisions
        self.assertEqual({revision_id}, set(repo.all_revision_ids()))
        # and can also query the parent map, either on the revisions
        # versionedfiles, which works in tuple keys...
        self.assertEqual({(revision_id,): ()},
                         repo.revisions.get_parent_map([(revision_id,)]))
        # ... or on the repository directly...
        self.assertEqual({revision_id: (NULL_REVISION,)},
                         repo.get_parent_map([revision_id]))
        # ... or on the repository's graph.
        self.assertEqual({revision_id: (NULL_REVISION,)},
                         repo.get_graph().get_parent_map([revision_id]))
        # ... or on the repository's graph, when there is an other repository.
        other = self.make_repository('other')
        other.lock_read()
        self.addCleanup(other.unlock)
        self.assertEqual({revision_id: (NULL_REVISION,)},
                         repo.get_graph(other).get_parent_map([revision_id]))



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.pyutils."""

from breezy import branch, tests
from breezy.pyutils import calc_parent_name, get_named_object


class TestGetNamedObject(tests.TestCase):
    """Tests for get_named_object."""

    def test_module_only(self):
        import sys
        self.assertIs(sys, get_named_object('sys'))

    def test_dotted_module(self):
        self.assertIs(branch, get_named_object('breezy.branch'))

    def test_module_attr(self):
        self.assertIs(
            branch.Branch, get_named_object('breezy.branch', 'Branch'))

    def test_dotted_attr(self):
        self.assertIs(
            branch.Branch.hooks,
            get_named_object('breezy.branch', 'Branch.hooks'))

    def test_package(self):
        # breezy.tests is a package, not simply a module
        self.assertIs(tests, get_named_object('breezy.tests'))

    def test_package_attr(self):
        # breezy.tests is a package, not simply a module
        self.assertIs(
            tests.TestCase, get_named_object('breezy.tests', 'TestCase'))

    def test_import_error(self):
        self.assertRaises(ImportError, get_named_object, 'NO_SUCH_MODULE')

    def test_attribute_error(self):
        self.assertRaises(
            AttributeError, get_named_object, 'sys', 'NO_SUCH_ATTR')


class TestCalcParent_name(tests.TestCase):
    """Tests for calc_parent_name."""

    def test_dotted_member(self):
        self.assertEqual(
            ('mod_name', 'attr1', 'attr2'),
            calc_parent_name('mod_name', 'attr1.attr2'))

    def test_undotted_member(self):
        self.assertEqual(
            ('mod_name', None, 'attr1'),
            calc_parent_name('mod_name', 'attr1'))

    def test_dotted_module_no_member(self):
        self.assertEqual(
            ('mod', None, 'sub_mod'),
            calc_parent_name('mod.sub_mod'))

    def test_undotted_module_no_member(self):
        err = self.assertRaises(AssertionError, calc_parent_name, 'mod_name')
        self.assertEqual(
            "No parent object for top-level module 'mod_name'", err.args[0])



# Copyright (C) 2008, 2009, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Views stored within a working tree.

The views are actually in the WorkingTree.views namespace, but these are
1:1 with WorkingTree implementations so can be tested from here.
"""


from breezy import views as _mod_views
from breezy.tests import TestNotApplicable, TestSkipped
from breezy.tests.per_workingtree import TestCaseWithWorkingTree
from breezy.workingtree import WorkingTree


class TestTreeViews(TestCaseWithWorkingTree):

    def setUp(self):
        # formats that don't support views can skip the rest of these
        # tests...
        fmt = self.workingtree_format
        f = getattr(fmt, 'supports_views')
        if f is None:
            raise TestSkipped("format %s doesn't declare whether it "
                              "supports views, assuming not" % fmt)
        if not f():
            raise TestNotApplicable("format %s doesn't support views" % fmt)
        super().setUp()

    def test_views_initially_empty(self):
        wt = self.make_branch_and_tree('wt')
        current, views = wt.views.get_view_info()
        self.assertEqual(None, current)
        self.assertEqual({}, views)

    def test_set_and_get_view_info(self):
        wt = self.make_branch_and_tree('wt')
        view_current = 'view-name'
        view_dict = {
            view_current: ['dir-1'],
            'other-name': ['dir-2']}
        wt.views.set_view_info(view_current, view_dict)
        current, views = wt.views.get_view_info()
        self.assertEqual(view_current, current)
        self.assertEqual(view_dict, views)
        # then reopen the tree and see they're still there
        wt = WorkingTree.open('wt')
        current, views = wt.views.get_view_info()
        self.assertEqual(view_current, current)
        self.assertEqual(view_dict, views)
        # test setting a current view which does not exist
        self.assertRaises(_mod_views.NoSuchView,
                          wt.views.set_view_info, 'yet-another', view_dict)
        current, views = wt.views.get_view_info()
        self.assertEqual(view_current, current)
        self.assertEqual(view_dict, views)
        # test clearing the current view
        wt.views.set_view_info(None, view_dict)
        current, views = wt.views.get_view_info()
        self.assertEqual(None, current)
        self.assertEqual(view_dict, views)

    def test_lookup_view(self):
        wt = self.make_branch_and_tree('wt')
        view_current = 'view-name'
        view_dict = {
            view_current: ['dir-1'],
            'other-name': ['dir-2']}
        wt.views.set_view_info(view_current, view_dict)
        # test lookup of the default view
        result = wt.views.lookup_view()
        self.assertEqual(result, ['dir-1'])
        # test lookup of a named view
        result = wt.views.lookup_view('other-name')
        self.assertEqual(result, ['dir-2'])

    def test_set_view(self):
        wt = self.make_branch_and_tree('wt')
        # test that set_view sets the current view by default
        wt.views.set_view('view-1', ['dir-1'])
        current, views = wt.views.get_view_info()
        self.assertEqual('view-1', current)
        self.assertEqual({'view-1': ['dir-1']}, views)
        # test adding a view and not making it the current one
        wt.views.set_view('view-2', ['dir-2'], make_current=False)
        current, views = wt.views.get_view_info()
        self.assertEqual('view-1', current)
        self.assertEqual({'view-1': ['dir-1'], 'view-2': ['dir-2']}, views)

    def test_unicode_view(self):
        wt = self.make_branch_and_tree('wt')
        view_name = '\u3070'
        view_files = ['foo', 'bar/']
        view_dict = {view_name: view_files}
        wt.views.set_view_info(view_name, view_dict)
        current, views = wt.views.get_view_info()
        self.assertEqual(view_name, current)
        self.assertEqual(view_dict, views)

    def test_no_such_view(self):
        wt = self.make_branch_and_tree('wt')
        try:
            wt.views.lookup_view('opaque')
        except _mod_views.NoSuchView as e:
            self.assertEqual(e.view_name, 'opaque')
            self.assertEqual(str(e), 'No such view: opaque.')
        else:
            self.fail("didn't get expected exception")

    def test_delete_view(self):
        wt = self.make_branch_and_tree('wt')
        view_name = '\N{GREEK SMALL LETTER ALPHA}'
        view_files = ['alphas/']
        wt.views.set_view(view_name, view_files)
        # now try to delete it
        wt.views.delete_view(view_name)
        # now you can't look it up
        self.assertRaises(_mod_views.NoSuchView,
                          wt.views.lookup_view, view_name)
        # and it's not in the dictionary
        self.assertEqual(wt.views.get_view_info()[1], {})
        # and you can't remove it a second time
        self.assertRaises(_mod_views.NoSuchView,
                          wt.views.delete_view, view_name)
        # or remove a view that never existed
        self.assertRaises(_mod_views.NoSuchView,
                          wt.views.delete_view, view_name + '2')

    def test_check_path_in_view(self):
        wt = self.make_branch_and_tree('wt')
        view_current = 'view-name'
        view_dict = {
            view_current: ['dir-1'],
            'other-name': ['dir-2']}
        wt.views.set_view_info(view_current, view_dict)
        self.assertEqual(_mod_views.check_path_in_view(wt, 'dir-1'), None)
        self.assertEqual(_mod_views.check_path_in_view(wt, 'dir-1/sub'), None)
        self.assertRaises(_mod_views.FileOutsideView,
                          _mod_views.check_path_in_view, wt, 'dir-2')
        self.assertRaises(_mod_views.FileOutsideView,
                          _mod_views.check_path_in_view, wt, 'dir-2/sub')
        self.assertRaises(_mod_views.FileOutsideView,
                          _mod_views.check_path_in_view, wt, 'other')


class TestUnsupportedViews(TestCaseWithWorkingTree):
    """Formats that don't support views should give reasonable errors."""

    def setUp(self):
        fmt = self.workingtree_format
        supported = getattr(fmt, 'supports_views')
        if supported is None:
            warn("Format %s doesn't declare whether it supports views or not"
                 % fmt)
            raise TestSkipped('No view support at all')
        if supported():
            raise TestSkipped("Format %s declares that views are supported"
                              % fmt)
            # it's covered by TestTreeViews
        super().setUp()

    def test_view_methods_raise(self):
        wt = self.make_branch_and_tree('wt')
        self.assertRaises(_mod_views.ViewsNotSupported,
                          wt.views.set_view_info, 'bar', {'bar': ['bars/']})
        self.assertRaises(_mod_views.ViewsNotSupported,
                          wt.views.get_view_info)
        self.assertRaises(_mod_views.ViewsNotSupported,
                          wt.views.lookup_view, 'foo')
        self.assertRaises(_mod_views.ViewsNotSupported,
                          wt.views.set_view, 'foo', 'bar')
        self.assertRaises(_mod_views.ViewsNotSupported,
                          wt.views.delete_view, 'foo')



# Copyright (C) 2009-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests from fetching from git into bzr."""

import os
import stat
import time

from dulwich.objects import S_IFGITLINK, Blob, Tag, Tree
from dulwich.repo import Repo as GitRepo

from ... import osutils
from ...branch import Branch
from ...bzr import knit, versionedfile
from ...bzr.inventory import Inventory
from ...controldir import ControlDir
from ...repository import Repository
from ...tests import TestCaseWithTransport
from ..fetch import import_git_blob, import_git_submodule, import_git_tree
from ..mapping import DEFAULT_FILE_MODE, BzrGitMappingv1
from . import GitBranchBuilder


class RepositoryFetchTests:

    def make_git_repo(self, path):
        os.mkdir(path)
        return GitRepo.init(os.path.abspath(path))

    def clone_git_repo(self, from_url, to_url, revision_id=None):
        oldrepos = self.open_git_repo(from_url)
        dir = ControlDir.create(to_url)
        newrepos = dir.create_repository()
        oldrepos.copy_content_into(newrepos, revision_id=revision_id)
        return newrepos

    def test_empty(self):
        self.make_git_repo("d")
        newrepos = self.clone_git_repo("d", "f")
        self.assertEqual([], newrepos.all_revision_ids())

    def make_onerev_branch(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", b"foo\nbar\n", False)
        mark = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg")
        gitsha = bb.finish()[mark]
        os.chdir("..")
        return "d", gitsha

    def test_single_rev(self):
        path, gitsha = self.make_onerev_branch()
        oldrepo = self.open_git_repo(path)
        newrepo = self.clone_git_repo(path, "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        self.assertEqual([revid], newrepo.all_revision_ids())

    def test_single_rev_specific(self):
        path, gitsha = self.make_onerev_branch()
        oldrepo = self.open_git_repo(path)
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        newrepo = self.clone_git_repo(path, "f", revision_id=revid)
        self.assertEqual([revid], newrepo.all_revision_ids())

    def test_incremental(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", b"foo\nbar\n", False)
        mark1 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg")
        bb.set_file("foobar", b"fooll\nbar\n", False)
        mark2 = bb.commit(b"Somebody <somebody@someorg.org>", b"nextmsg")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        newrepo = self.clone_git_repo("d", "f", revision_id=revid1)
        self.assertEqual([revid1], newrepo.all_revision_ids())
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        newrepo.fetch(oldrepo, revision_id=revid2)
        self.assertEqual({revid1, revid2},
                         set(newrepo.all_revision_ids()))

    def test_dir_becomes_symlink(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("mylink/somefile", b"foo\nbar\n", False)
        mark1 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg1")
        bb.delete_entry("mylink/somefile")
        bb.set_symlink("mylink", "target/")
        mark2 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg2")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree1 = newrepo.revision_tree(revid1)
        tree2 = newrepo.revision_tree(revid2)
        self.assertEqual(revid1, tree1.get_file_revision("mylink"))
        self.assertEqual("directory", tree1.kind("mylink"))
        self.assertEqual(None, tree1.get_symlink_target("mylink"))
        self.assertEqual(revid2, tree2.get_file_revision("mylink"))
        self.assertEqual("symlink", tree2.kind("mylink"))
        self.assertEqual("target/", tree2.get_symlink_target("mylink"))

    def test_symlink_becomes_dir(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_symlink("mylink", "target/")
        mark1 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg1")
        bb.delete_entry("mylink")
        bb.set_file("mylink/somefile", b"foo\nbar\n", False)
        mark2 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg2")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree1 = newrepo.revision_tree(revid1)
        tree2 = newrepo.revision_tree(revid2)
        self.assertEqual(revid1, tree1.get_file_revision("mylink"))
        self.assertEqual("symlink", tree1.kind("mylink"))
        self.assertEqual("target/", tree1.get_symlink_target("mylink"))
        self.assertEqual(revid2, tree2.get_file_revision("mylink"))
        self.assertEqual("directory", tree2.kind("mylink"))
        self.assertEqual(None, tree2.get_symlink_target("mylink"))

    def test_changing_symlink(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_symlink("mylink", "target")
        mark1 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg1")
        bb.set_symlink("mylink", "target/")
        mark2 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg2")
        marks = bb.finish()
        gitsha1 = marks[mark1]
        gitsha2 = marks[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree1 = newrepo.revision_tree(revid1)
        tree2 = newrepo.revision_tree(revid2)
        self.assertEqual(revid1, tree1.get_file_revision("mylink"))
        self.assertEqual("target", tree1.get_symlink_target("mylink"))
        self.assertEqual(revid2, tree2.get_file_revision("mylink"))
        self.assertEqual("target/", tree2.get_symlink_target("mylink"))

    def test_executable(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", b"foo\nbar\n", True)
        bb.set_file("notexec", b"foo\nbar\n", False)
        mark = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg")
        gitsha = bb.finish()[mark]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        tree = newrepo.revision_tree(revid)
        self.assertTrue(tree.has_filename("foobar"))
        self.assertEqual(True, tree.is_executable("foobar"))
        self.assertTrue(tree.has_filename("notexec"))
        self.assertEqual(False, tree.is_executable("notexec"))

    def test_becomes_executable(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", b"foo\nbar\n", False)
        mark1 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg")
        bb.set_file("foobar", b"foo\nbar\n", True)
        mark2 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg")
        gitsha2 = bb.finish()[mark2]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        tree = newrepo.revision_tree(revid)
        self.assertTrue(tree.has_filename("foobar"))
        self.assertEqual(True, tree.is_executable("foobar"))
        self.assertEqual(revid, tree.get_file_revision("foobar"))

    def test_into_stacked_on(self):
        r = self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", b"foo\n", False)
        mark1 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg1")
        gitsha1 = bb.finish()[mark1]
        os.chdir("..")
        stacked_on = self.clone_git_repo("d", "stacked-on")
        oldrepo = Repository.open("d")
        revid1 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha1)
        self.assertEqual([revid1], stacked_on.all_revision_ids())
        b = stacked_on.controldir.create_branch()
        b.generate_revision_history(revid1)
        self.assertEqual(b.last_revision(), revid1)
        tree = self.make_branch_and_tree("stacked")
        tree.branch.set_stacked_on_url(b.user_url)
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("barbar", b"bar\n", False)
        bb.set_file("foo/blie/bla", b"bla\n", False)
        mark2 = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg2")
        gitsha2 = bb.finish()[mark2]
        revid2 = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha2)
        os.chdir("..")
        tree.branch.fetch(Branch.open("d"))
        tree.branch.repository.check()
        self.addCleanup(tree.lock_read().unlock)
        self.assertEqual(
            {(revid2,)},
            tree.branch.repository.revisions.without_fallbacks().keys())
        self.assertEqual(
            {revid1, revid2},
            set(tree.branch.repository.all_revision_ids()))

    def test_non_ascii_characters(self):
        self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foőbar", b"foo\nbar\n", False)
        mark = bb.commit(b"Somebody <somebody@someorg.org>", b"mymsg")
        gitsha = bb.finish()[mark]
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        newrepo = self.clone_git_repo("d", "f")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        tree = newrepo.revision_tree(revid)
        self.assertTrue(tree.has_filename("foőbar"))

    def test_tagged_tree(self):
        r = self.make_git_repo("d")
        os.chdir("d")
        bb = GitBranchBuilder()
        bb.set_file("foobar", b"fooll\nbar\n", False)
        mark = bb.commit(b"Somebody <somebody@someorg.org>", b"nextmsg")
        marks = bb.finish()
        gitsha = marks[mark]
        tag = Tag()
        tag.name = b"sometag"
        tag.tag_time = int(time.time())
        tag.tag_timezone = 0
        tag.tagger = b"Somebody <somebody@example.com>"
        tag.message = b"Created tag pointed at tree"
        tag.object = (Tree, r[gitsha].tree)
        r.object_store.add_object(tag)
        r[b"refs/tags/sometag"] = tag
        os.chdir("..")
        oldrepo = self.open_git_repo("d")
        revid = oldrepo.get_mapping().revision_id_foreign_to_bzr(gitsha)
        newrepo = self.clone_git_repo("d", "f")
        self.assertEqual({revid}, set(newrepo.all_revision_ids()))


class LocalRepositoryFetchTests(RepositoryFetchTests, TestCaseWithTransport):

    def open_git_repo(self, path):
        return Repository.open(path)


class DummyStoreUpdater:

    def add_object(self, obj, ie, path):
        pass

    def finish(self):
        pass


class ImportObjects(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self._mapping = BzrGitMappingv1()
        factory = knit.make_file_factory(True, versionedfile.PrefixMapper())
        self._texts = factory(self.get_transport('texts'))

    def test_import_blob_missing_in_one_parent(self):
        builder = self.make_branch_builder('br')
        builder.start_series()
        rev_root = builder.build_snapshot(None, [
            ('add', ('', b'rootid', 'directory', ''))])
        rev1 = builder.build_snapshot([rev_root], [
            ('add', ('bla', self._mapping.generate_file_id('bla'), 'file', b'content'))])
        rev2 = builder.build_snapshot([rev_root], [])
        builder.finish_series()
        branch = builder.get_branch()

        blob = Blob.from_string(b"bar")
        objs = {"blobname": blob}
        ret = import_git_blob(self._texts, self._mapping, b"bla", b"bla",
                              (None, "blobname"),
                              branch.repository.revision_tree(
                                  rev1), b'rootid', b"somerevid",
                              [branch.repository.revision_tree(r) for r in [
                                  rev1, rev2]],
                              objs.__getitem__,
                              (None, DEFAULT_FILE_MODE), DummyStoreUpdater(),
                              self._mapping.generate_file_id)
        self.assertEqual({(b'git:bla', b'somerevid')}, self._texts.keys())

    def test_import_blob_simple(self):
        blob = Blob.from_string(b"bar")
        objs = {"blobname": blob}
        ret = import_git_blob(self._texts, self._mapping, b"bla", b"bla",
                              (None, "blobname"),
                              None, None, b"somerevid", [], objs.__getitem__,
                              (None, DEFAULT_FILE_MODE), DummyStoreUpdater(),
                              self._mapping.generate_file_id)
        self.assertEqual({(b'git:bla', b'somerevid')}, self._texts.keys())
        self.assertEqual(next(self._texts.get_record_stream([(b'git:bla', b'somerevid')],
                                                            "unordered", True)).get_bytes_as("fulltext"), b"bar")
        self.assertEqual(1, len(ret))
        self.assertEqual(None, ret[0][0])
        self.assertEqual("bla", ret[0][1])
        ie = ret[0][3]
        self.assertEqual(False, ie.executable)
        self.assertEqual("file", ie.kind)
        self.assertEqual(b"somerevid", ie.revision)
        self.assertEqual(osutils.sha_strings([b"bar"]), ie.text_sha1)

    def test_import_tree_empty_root(self):
        tree = Tree()
        ret, child_modes = import_git_tree(self._texts, self._mapping, b"", b"",
                                           (None, tree.id), None,
                                           None, b"somerevid", [], {
                                               tree.id: tree}.__getitem__,
                                           (None, stat.S_IFDIR), DummyStoreUpdater(),
                                           self._mapping.generate_file_id)
        self.assertEqual(child_modes, {})
        self.assertEqual(
            {(b"TREE_ROOT", b'somerevid')}, self._texts.keys())
        self.assertEqual(1, len(ret))
        self.assertEqual(None, ret[0][0])
        self.assertEqual("", ret[0][1])
        ie = ret[0][3]
        self.assertEqual(False, ie.executable)
        self.assertEqual("directory", ie.kind)
        self.assertEqual({}, ie.children)
        self.assertEqual(b"somerevid", ie.revision)
        self.assertEqual(None, ie.text_sha1)

    def test_import_tree_empty(self):
        tree = Tree()
        ret, child_modes = import_git_tree(self._texts, self._mapping, b"bla", b"bla",
                                           (None, tree.id), None, None, b"somerevid", [],
                                           {tree.id: tree}.__getitem__,
                                           (None, stat.S_IFDIR), DummyStoreUpdater(),
                                           self._mapping.generate_file_id)
        self.assertEqual(child_modes, {})
        self.assertEqual({(b"git:bla", b'somerevid')}, self._texts.keys())
        self.assertEqual(1, len(ret))
        self.assertEqual(None, ret[0][0])
        self.assertEqual("bla", ret[0][1])
        ie = ret[0][3]
        self.assertEqual("directory", ie.kind)
        self.assertEqual(False, ie.executable)
        self.assertEqual({}, ie.children)
        self.assertEqual(b"somerevid", ie.revision)
        self.assertEqual(None, ie.text_sha1)

    def test_import_tree_with_file(self):
        blob = Blob.from_string(b"bar1")
        tree = Tree()
        tree.add(b"foo", stat.S_IFREG | 0o644, blob.id)
        objects = {blob.id: blob, tree.id: tree}
        ret, child_modes = import_git_tree(self._texts, self._mapping, b"bla", b"bla",
                                           (None, tree.id), None, None, b"somerevid", [],
                                           objects.__getitem__, (None, stat.S_IFDIR), DummyStoreUpdater(
                                           ),
                                           self._mapping.generate_file_id)
        self.assertEqual(child_modes, {})
        self.assertEqual(2, len(ret))
        self.assertEqual(None, ret[0][0])
        self.assertEqual("bla", ret[0][1])
        self.assertEqual(None, ret[1][0])
        self.assertEqual("bla/foo", ret[1][1])
        ie = ret[0][3]
        self.assertEqual("directory", ie.kind)
        ie = ret[1][3]
        self.assertEqual("file", ie.kind)
        self.assertEqual(b"git:bla/foo", ie.file_id)
        self.assertEqual(b"somerevid", ie.revision)
        self.assertEqual(osutils.sha_strings([b"bar1"]), ie.text_sha1)
        self.assertEqual(False, ie.executable)

    def test_import_tree_with_unusual_mode_file(self):
        blob = Blob.from_string(b"bar1")
        tree = Tree()
        tree.add(b"foo", stat.S_IFREG | 0o664, blob.id)
        objects = {blob.id: blob, tree.id: tree}
        ret, child_modes = import_git_tree(self._texts, self._mapping,
                                           b"bla", b"bla", (None, tree.id), None, None, b"somerevid", [
                                               ],
                                           objects.__getitem__, (None, stat.S_IFDIR), DummyStoreUpdater(
                                           ),
                                           self._mapping.generate_file_id)
        self.assertEqual(child_modes, {b"bla/foo": stat.S_IFREG | 0o664})

    def test_import_tree_with_file_exe(self):
        blob = Blob.from_string(b"bar")
        tree = Tree()
        tree.add(b"foo", 0o100755, blob.id)
        objects = {blob.id: blob, tree.id: tree}
        ret, child_modes = import_git_tree(self._texts, self._mapping, b"", b"",
                                           (None, tree.id), None, None, b"somerevid", [],
                                           objects.__getitem__, (None, stat.S_IFDIR), DummyStoreUpdater(
                                           ),
                                           self._mapping.generate_file_id)
        self.assertEqual(child_modes, {})
        self.assertEqual(2, len(ret))
        self.assertEqual(None, ret[0][0])
        self.assertEqual("", ret[0][1])
        self.assertEqual(None, ret[1][0])
        self.assertEqual("foo", ret[1][1])
        ie = ret[0][3]
        self.assertEqual("directory", ie.kind)
        ie = ret[1][3]
        self.assertEqual("file", ie.kind)
        self.assertEqual(True, ie.executable)

    def test_directory_converted_to_submodule(self):
        base_inv = Inventory()
        base_inv.add_path("foo", "directory")
        base_inv.add_path("foo/bar", "file")
        othertree = Blob.from_string(b"someotherthing")
        blob = Blob.from_string(b"bar")
        tree = Tree()
        tree.add(b"bar", 0o160000, blob.id)
        objects = {tree.id: tree}
        ret, child_modes = import_git_submodule(self._texts, self._mapping, b"foo", b"foo",
                                                (tree.id, othertree.id), base_inv, base_inv.root.file_id, b"somerevid", [
                                                    ],
                                                objects.__getitem__, (stat.S_IFDIR | 0o755, S_IFGITLINK), DummyStoreUpdater(
                                                ),
                                                self._mapping.generate_file_id)
        self.assertEqual(child_modes, {})
        self.assertEqual(2, len(ret))
        self.assertEqual(
            ret[0], ("foo/bar", None, base_inv.path2id("foo/bar"), None))
        self.assertEqual(ret[1][:3], ("foo", "foo",
                                      self._mapping.generate_file_id("foo")))
        ie = ret[1][3]
        self.assertEqual(ie.kind, "tree-reference")
